"""RH56DFTP Hand module for Inspire finger joints."""

from .rh56dftp_hand import InspireFingerJoint, create_finger_joints_from_config

__all__ = ['InspireFingerJoint', 'create_finger_joints_from_config']




from ctypes import *
from socket import CAN_EFF_FLAG, CAN_EFF_MASK
from feathersdk.utils.feathertypes import Callable, Any, Type, Optional, Self, Union
import os
import copy
from enum import Enum
from .recompilation import recompile_if_needed, _LIB_SO
from feathersdk.utils.common import timestamped_class

# Info for loading the library
_LIB = None

# Make sure the main callback function is not garbage collected while the C polling loop is running.
_MAIN_CALLBACK_FUNC = None

# C lib structure info
_SOCKET_NAME_MAX_LEN = 32
_CAN_DATA_SIZE = 8
_TCP_BUF_SIZE = 512
_MAX_ERR_STR_LEN = 1024
_POLL_TIMEOUT_MS = 1

class SocketType(Enum):
    UNINITIALIZED = 0
    CAN = 1
    TCP = 2
    ERROR = 3

class SCTSystemError(Exception):
    pass

class SCTOverflowError(Exception):
    pass

class SCTUserError(Exception):
    pass

class SCTTimeoutError(Exception):
    pass

class SCTErrorType(Enum):
    NO_OVERRIDE = -1
    NONE = 0
    SYSTEM = 1
    OVERFLOW = 2
    USER = 3
    TIMEOUT = 4

    @classmethod
    def raise_error(cls, err_type: int, err_msg: str) -> None:
        if err_type == cls.NO_OVERRIDE.value:
            raise ValueError("Should not be able to get NO_OVERRIDE error type!")
        elif err_type == cls.NONE.value:
            raise ValueError("No error occurred! Error message: " + err_msg)
        elif err_type == cls.SYSTEM.value:
            raise SCTSystemError(err_msg)
        elif err_type == cls.OVERFLOW.value:
            raise SCTOverflowError(err_msg)
        elif err_type == cls.USER.value:
            raise SCTUserError(err_msg)
        elif err_type == cls.TIMEOUT.value:
            raise SCTTimeoutError(err_msg)
        else:
            raise ValueError(f"Unhandled error type: {err_type}")


def _to_c_array(data: Union[bytes, str], length: int) -> 'Array[c_uint8]':
    """Convert a Python string or bytes object to a ctypes array of uint8_t."""
    if isinstance(data, bytes):
        return (c_uint8 * length)(*data)
    elif isinstance(data, str):
        return (c_uint8 * length)(*data.encode("ascii"))
    else:
        raise ValueError(f"Unhandled data type: {type(data).__name__}")


@timestamped_class
class SocketResult(Structure):
    """Result of a socket operation. Can be a CAN message, a TCP message, or an error."""

    _fields_ = [
        ("result_type", c_uint8),
        ("socket_name", c_uint8 * _SOCKET_NAME_MAX_LEN),

        ("can_id", c_uint32),
        ("data", c_uint8 * _CAN_DATA_SIZE),
        ("dlc", c_uint8),

        ("tcp_buffer", c_uint8 * _TCP_BUF_SIZE),
        ("tcp_buf_len", c_uint32),

        ("err_type", c_int32),
        ("err_msg", c_uint8 * _MAX_ERR_STR_LEN),
    ]

    @classmethod
    def new(cls: Type[Self], result_type: SocketType, socket_name: str, can_id: Optional[int] = None, 
            data: Optional[bytes] = None, dlc: Optional[int] = None, tcp_buffer: Optional[bytes] = None, 
            tcp_buf_len: Optional[int] = None, err_type: Optional[SCTErrorType] = None, 
            err_msg: Optional[str] = None) -> Self:
        
        socket_name = _to_c_array(socket_name, _SOCKET_NAME_MAX_LEN)
        
        if result_type == SocketType.CAN:
            if can_id is None or data is None or dlc is None:
                raise ValueError("can_id, data, and dlc are required for CAN messages")
            buf = _to_c_array(data, _CAN_DATA_SIZE)
            return cls(result_type=result_type.value, socket_name=socket_name, can_id=can_id, data=buf, dlc=dlc)
        elif result_type == SocketType.TCP:
            if tcp_buffer is None or tcp_buf_len is None:
                raise ValueError("tcp_buffer and tcp_buf_len are required for TCP messages")
            buf = _to_c_array(tcp_buffer, _TCP_BUF_SIZE)
            return cls(result_type=result_type.value, socket_name=socket_name, tcp_buffer=buf, tcp_buf_len=tcp_buf_len)
        elif result_type == SocketType.ERROR:
            if err_type is None or err_msg is None:
                raise ValueError("err_type and err_msg are required for ERROR messages")
            buf = _to_c_array(err_msg, _MAX_ERR_STR_LEN)
            return cls(result_type=result_type.value, socket_name=socket_name, err_type=err_type.value, err_msg=buf)
        else:
            raise NotImplementedError(f"Unhandled result type: {result_type}")

    # Custom getter for uint8_t array fields to return bytes/strings instead of ctypes array objects.
    def __getattribute__(self, name: str):
        ret = super().__getattribute__(name)

        if name == "socket_name" or name == "err_msg":
            return bytes(ret).partition(b"\0")[0].decode("ascii")
        elif type(type(ret)) is _C_ARRAY_TYPE:
            if ret._type_ is c_ubyte:
                return bytes(ret)
            else:
                raise TypeError(f"Unhandled array type: {type(ret).__name__}")
            
        return ret

    def __repr__(self) -> str:
        if self.result_type == SocketType.UNINITIALIZED.value:
            return "SocketResult(result_type=SocketType.UNINITIALIZED)"
        elif self.result_type == SocketType.CAN.value:
            # Format can id as 3 hex digits if standard (11-bit), 8 hex digits if extended (29-bit)
            can_id_str = f"0x{self.can_id & 0x7FF:03X}" if not self.is_extended_can() else f"0x{self.can_id:08X}"
            can_id_str = f"0x{self.can_id:08X}"
            return f"SocketResult(result_type=SocketType.CAN, socket_name={self.socket_name}, " \
                f"can_id={can_id_str}, data={self.data}, dlc={self.dlc})"
        elif self.result_type == SocketType.TCP.value:
            return f"SocketResult(result_type=SocketType.TCP, socket_name={self.socket_name}, " \
                f"tcp_buffer={self.tcp_buffer}, tcp_buf_len={self.tcp_buf_len})"
        elif self.result_type == SocketType.ERROR.value:
            return f"SocketResult(result_type=SocketType.ERROR, err_type={self.err_type}, err_msg={self.err_msg})"
        else:
            raise NotImplementedError(f"Unhandled result type: {self.result_type}")
    
    def __str__(self) -> str:
        return self.__repr__()
    
    def __hash__(self) -> int:
        if self.result_type == SocketType.CAN.value:
            return hash((self.result_type, self.can_id, self.data, self.dlc))
        elif self.result_type == SocketType.TCP.value:
            return hash((self.result_type, self.tcp_buffer, self.tcp_buf_len))
        elif self.result_type == SocketType.ERROR.value:
            return hash((self.result_type, self.err_type, self.err_msg))
        else:
            raise NotImplementedError(f"Unhandled result type: {self.result_type}")
    
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SocketResult) or self.result_type != other.result_type:
            return False
        if self.result_type == SocketType.CAN.value:
            return self.can_id == other.can_id and self.data == other.data and self.dlc == other.dlc
        elif self.result_type == SocketType.TCP.value:
            return self.tcp_buffer == other.tcp_buffer and self.tcp_buf_len == other.tcp_buf_len
        elif self.result_type == SocketType.ERROR.value:
            return self.err_type == other.err_type and self.err_msg == other.err_msg
        else:
            raise NotImplementedError(f"Unhandled result type: {self.result_type}")
    
    def is_error(self) -> bool:
        """Check if the result is an error."""
        return self.result_type == SocketType.ERROR.value
    
    def is_can(self) -> bool:
        """Check if the result is a CAN message."""
        return self.result_type == SocketType.CAN.value
    
    def is_tcp(self) -> bool:
        """Check if the result is a TCP message."""
        return self.result_type == SocketType.TCP.value
    
    def is_extended_can(self) -> bool:
        """Check if the result is an extended CAN message."""
        return (self.can_id & CAN_EFF_FLAG) != 0
    
    @property
    def can_id_no_flags(self) -> int:
        """Get the CAN ID without the can protocol flags (set first 3 bits to 0)"""
        return self.can_id & CAN_EFF_MASK


_MAIN_CALLBACK_TYPE = CFUNCTYPE(None, POINTER(SocketResult))


def __compile_load_lib() -> None:
    """Load the shared library, attempting to compile it if it doesn't exist. Should only be called in this file."""
    global _LIB
    
    # Load the library, and set up prototypes.
    _LIB = CDLL(recompile_if_needed())

    _LIB.get_err_str.argtypes = [c_char_p]
    _LIB.get_err_str.restype = None

    _LIB.get_err_type.argtypes = None
    _LIB.get_err_type.restype = c_int32

    _LIB.clear_err_str.argtypes = None
    _LIB.clear_err_str.restype = None

    _LIB.cansend.argtypes = [c_char_p, c_bool, c_uint32, c_uint8 * _CAN_DATA_SIZE, c_uint8]
    _LIB.cansend.restype = c_int

    _LIB.tcpsend_modbus.argtypes = [c_char_p, c_uint16, c_uint16, c_uint8, c_uint16, c_uint16]
    _LIB.tcpsend_modbus.restype = c_int

    _LIB.clear_all_sockets.argtypes = None
    _LIB.clear_all_sockets.restype = c_int

    _LIB.remove_endpoint.argtypes = [c_char_p]
    _LIB.remove_endpoint.restype = c_int

    _LIB.add_endpoint.argtypes = [c_char_p]
    _LIB.add_endpoint.restype = c_int

    _LIB.polling_step.argtypes = [_MAIN_CALLBACK_TYPE, c_int]
    _LIB.polling_step.restype = c_int


def _unload_socketcan_library(delete_lib: bool = False) -> None:
    """Unload the socketcan library, deleting the library .so file if requested.
    
    Mostly just used for testing purposes.

    Args:
        delete_lib: If True, will delete the library .so file after unloading it.
    """
    global _LIB
    _LIB = None
    if delete_lib and os.path.exists(_LIB_SO):
        os.remove(_LIB_SO)
    
    # Force the gc to run. Will unload the library so long as it has no more references.
    import gc
    gc.collect()


# All of the main super-types for C types. All c-types are derivatives of these.
_C_SIMPLE_TYPE = type(c_uint8)
_C_ARRAY_TYPE = type(c_uint8 * 2)
_C_POINTER_TYPE = type(POINTER(c_uint8))
_C_STRUCT_TYPE = type(Structure)
_C_FUNC_TYPE = type(CFUNCTYPE(None, c_uint8))
_C_SUPER_TYPES = set([_C_SIMPLE_TYPE, _C_ARRAY_TYPE, _C_POINTER_TYPE, _C_STRUCT_TYPE, _C_FUNC_TYPE])

# Bounds for the integer types in the library (since python doesn't have native integer bounds). Inclusive on both ends.
_SINT_TS = [c_int, c_int16, c_int32, c_int64, c_int8, c_long, c_longdouble, c_longlong, c_short, c_byte, c_char]
_UINT_TS = [c_size_t, c_ssize_t, c_ubyte, c_uint, c_uint16, c_uint32, c_uint64, c_uint8, c_ulong, c_ulonglong, c_ushort]
_INT_BOUNDS = {
    **{t: (-2**(8*sizeof(t)-1), 2**(8*sizeof(t)-1)-1) for t in _SINT_TS},
    **{t: (0, 2**(8*sizeof(t))-1) for t in _UINT_TS},
}

# Types that don't need validation.
_UNVALIDATED_SUPER_TYPES = set([_C_POINTER_TYPE, _C_STRUCT_TYPE, _C_FUNC_TYPE])
_UNVALIDATED_TYPES = set([c_void_p, c_char_p, c_wchar_p])


def __get_lib() -> CDLL:
    """Get the shared library, loading it if it isn't already loaded. Should only be called in this file."""
    global _LIB
    if _LIB is None:
        __compile_load_lib()
    return _LIB

def __check_for_error(result: int) -> None:
    """Check for an error in the result and raise the appropriate error."""
    if result != 0:
        err_type, err_str = _get_err_type(), _get_err_str()
        _clear_err_str()
        SCTErrorType.raise_error(err_type, err_str)

def __validate_simple_arg(arg: Any, arg_type: type) -> None:
    """Validate a simple argument."""
    if arg_type in _INT_BOUNDS:
        min_val, max_val = _INT_BOUNDS[arg_type]
        if not (min_val <= arg <= max_val):
            raise OverflowError(f"Argument {arg} out of bounds for type {arg_type.__name__}: [{min_val}, {max_val}]")
    elif arg_type is c_bool:
        if arg is not True and arg is not False:
            raise TypeError(f"Expected boolean, got {arg}")
    elif arg_type is c_float or arg_type is c_double:
        if arg_type(arg).value == float("inf"):
            raise OverflowError(f"Argument {arg} is infinity or too large to store in a {arg_type.__name__}")
    elif arg_type not in _UNVALIDATED_TYPES:
        raise ValueError(f"Unhandled simple C type: {arg_type.__name__}")

def __validate_and_call(func: Callable[[Any], Any], *args: Any) -> Any:
    """Validate the arguments and call the library function, returning the result
    
    Does some validation to make sure the arguments will work with the library function. Currently:
    - Ensures integer overflow/underflow will raise an error.
    - Ensured floats/doubles are not infinity or too large to store in a C type (unless they are already C types).
    - Ensures boolean args are exact booleans like True or False.
    """
    args = list(args)
    if func.argtypes is None and len(args) > 0:
        raise ValueError(f"Expected no arguments, got {len(args)}")
    elif func.argtypes is not None and len(args) != len(func.argtypes):
        raise ValueError(f"Expected {len(func.argtypes)} lib arguments, got {len(args)}")

    for i, (arg, arg_type) in enumerate(zip(args, func.argtypes or [])):
        # If the arg is already a C type, don't validate it.
        if type(type(arg)) in _C_SUPER_TYPES:
            continue
        
        # Simple and array types can be handled
        elif type(arg_type) is _C_SIMPLE_TYPE:
            __validate_simple_arg(arg, arg_type)
        elif type(arg_type) is _C_ARRAY_TYPE:
            if len(arg) != arg_type._length_:
                raise ValueError(f"Invalid length for array {arg_type.__name__}: {len(arg)} != {arg_type._length_}")
            for item in arg:
                __validate_simple_arg(item, arg_type._type_)
            args[i] = arg_type(*arg)  # Convert the iterable of items to a ctypes array.

        # Other types like structures and functions are assumed to be validated already.
        elif type(arg_type) not in _UNVALIDATED_SUPER_TYPES:
            raise ValueError(f"Unhandled argument super-type: {type(arg_type).__name__}")
    
    return __check_for_error(func(*args))


def _get_err_str() -> str:
    """Get the current error string from the library."""
    err_str = create_string_buffer(_MAX_ERR_STR_LEN)
    __get_lib().get_err_str(err_str)
    return err_str.value.decode("ascii")

def _get_err_type() -> int:
    """Get the current error type from the library."""
    return __get_lib().get_err_type()

def _clear_err_str() -> None:
    """Clear the current error string from the library."""
    __get_lib().clear_err_str()

def _cansend(interface: str, extended: bool, can_id: int, data: bytes) -> None:
    """Send a CAN message to the given interface."""
    __validate_and_call(__get_lib().cansend, interface.encode("ascii"), extended, can_id, data, len(data))

def _tcpsend_modbus(ip: str, tid: int, uid: int, fcode: int, reg_addr: int, reg_val: int) -> None:
    """Send a Modbus TCP message to the given IP address."""
    __validate_and_call(__get_lib().tcpsend_modbus, ip.encode("ascii"), tid, uid, fcode, reg_addr, reg_val)

def _polling_loop(callback: Callable[[SocketResult], None]) -> None:
    """Subscribe to multiple endpoints and run infinite polling loop to call the callback for each message."""
    global _MAIN_CALLBACK_FUNC
    _MAIN_CALLBACK_FUNC = _MAIN_CALLBACK_TYPE(lambda result: callback(copy.deepcopy(result.contents)))
    __validate_and_call(__get_lib().polling_loop, _MAIN_CALLBACK_FUNC)

def _clear_all_sockets() -> None:
    """Clear all sockets."""
    __validate_and_call(__get_lib().clear_all_sockets)

def _remove_endpoint(endpoint: str) -> None:
    """Remove an endpoint."""
    __validate_and_call(__get_lib().remove_endpoint, endpoint.encode("ascii"))

def _add_endpoint(endpoint: str) -> None:
    """Add an endpoint."""
    __validate_and_call(__get_lib().add_endpoint, endpoint.encode("ascii"))

def _create_callback(callback: Callable[[SocketResult], None]) -> _MAIN_CALLBACK_TYPE:
    """Create a callback function for the library."""
    return _MAIN_CALLBACK_TYPE(lambda result: callback(copy.deepcopy(result.contents)))

def _polling_step(polling_callback: _MAIN_CALLBACK_TYPE, timeout_ms: int) -> None:
    """Step the polling loop for a given amount of time."""
    __validate_and_call(__get_lib().polling_step, polling_callback, timeout_ms)


__all__ = ["SocketType", "SocketResult", "SCTSystemError", "SCTOverflowError", "SCTUserError", "SCTTimeoutError", 
           "SCTErrorType"]


import time
from pathlib import Path

from .steppable_system import SteppableSystem
from abc import ABC, abstractmethod
from enum import Enum
from smbus2 import SMBus, i2c_msg
from feathersdk.comms.system import run_system_command
from feathersdk.utils.feathertypes import Optional, Tuple, Union
from feathersdk.utils.hydra_helper import load_hydra_config
from feathersdk.utils.logger import info
from dataclasses import dataclass, asdict
import traceback
import tempfile
import json
import os


MIN_VOLTAGE, MAX_VOLTAGE = 49.0, 55.0
ESTOP_VOLTAGE_THRESHOLD = 45
POWER_STATE_PATH = Path("/feathersys/state/power_state.json")
PICO_MAIN_PATH = Path(__file__).parent / "pico_main.py"

# Constants for I2C ADC
I2C_DEV_BUS = 1
ADS1115_ADDR = 0x48
CONFIG_REG = 0x01
CONV_REG = 0x00


def i2c_write(bus, addr, data):
    """Write data to I2C device.
    
    Args:
        bus: SMBus instance
        addr: I2C device address
        data: Data bytes to write
    """
    bus.i2c_rdwr(i2c_msg.write(addr, data))


def i2c_read(bus, addr, reg, length):
    """Read data from I2C device register.
    
    Args:
        bus: SMBus instance
        addr: I2C device address
        reg: Register address to read from
        length: Number of bytes to read
    
    Returns:
        list: List of bytes read from the register
    """
    write = i2c_msg.write(addr, [reg])
    read = i2c_msg.read(addr, length)
    bus.i2c_rdwr(write, read)
    return list(read)


@dataclass
class BatteryState:
    monitor_start_time: float
    last_estop_pressed_time: float
    last_estop_released_time: float
    last_reliable_seen_voltage: float
    state_update_time: float
    error: Optional[str] = None


def read_last_battery_state() -> BatteryState:
    text = "text wasn't read"
    for _ in range(5):
        try:
            text = POWER_STATE_PATH.read_text()
            if text is not None and text.strip() != "":
                break
        except (PermissionError, FileNotFoundError):
            pass
        time.sleep(0.005)
    try:
        return BatteryState(**json.loads(text))
    except:
        raise ValueError(f"Failed to parse battery state from {POWER_STATE_PATH}: \n\"{text}\"")


class PowerEvent(Enum):
    """Power system events."""
    POWER_OFF = 1
    POWER_RESTORED = 2
    POWER_UNHEALTHY = 3  # When we lose connection to the battery daemon


class PowerEventListener(ABC):
    """Abstract base class for power event listeners."""
    @abstractmethod
    def on_power_event(self, event: PowerEvent):
        """Handle power event notification.
        
        Args:
            event (PowerEvent): The power event that occurred
        """
        pass


def _get_battery_voltage_pico() -> Optional[float]:
    result = run_system_command(["/usr/bin/python", "-m", "mpremote", "run", PICO_MAIN_PATH], allow_error=True)
    if result is None or result.stdout is None or result.stdout.strip() == "":
        raise ValueError(f"Failed readingvoltage from Pico - got bad result or stdout: {result}, {result.stdout}")
    # voltage is computed battery voltage, adc_read is the raw ADC read value [0-3.3v]
    return list(map(float, result.stdout.strip().split()))[0]


def _get_battery_voltage_adc():
    """Get current battery voltage from I2C ADC.
    
    Reads voltage from ADS1115 ADC via I2C and applies scaling factors.
    
    Returns:
        float: Battery voltage in volts
    """
    with SMBus(I2C_DEV_BUS) as bus:
        # Configure ADS1115
        config = [
            CONFIG_REG,
            0xC0,  # AIN0, single-shot, PGA bits
            0x83   # 128 SPS, comparator disabled
        ]
        i2c_write(bus, ADS1115_ADDR, config)

        time.sleep(0.009)
        data = i2c_read(bus, ADS1115_ADDR, CONV_REG, 2)

        # Combine bytes (signed 16-bit)
        raw_adc = (data[0] << 8) | data[1]
        if raw_adc & 0x8000:
            raw_adc -= 0x10000

        # Convert to voltage (ADS1115 input) and apply resistor divider scaling (same as C code)
        return ((raw_adc * 6.114) / 32768.0) * 11


def get_battery_voltage(read_method: str) -> Union[float, str, None]:
    """Shouldn't call from feathersdk. Only used by battery daemon."""
    try:
        if read_method == "pico":
            ret_val = _get_battery_voltage_pico()
        elif read_method == "adc":
            ret_val = _get_battery_voltage_adc()
        else:
            raise ValueError(f"Invalid read method: {read_method}")
    except Exception as e:
        return traceback.format_exc()
    
    return None if ret_val is None or ret_val < 0.0 or ret_val > ESTOP_VOLTAGE_THRESHOLD * 1.5 else ret_val


class BatterySystem(SteppableSystem):
    """Base class for battery monitoring and power event management."""

    def __init__(self, config: dict = {}):
        super().__init__()
        self.operating_frequency = 10.0
        self.power_event_listeners = []
        self.last_state = read_last_battery_state()
        self.cfg = config
        self.rated_voltage = config.get("rated_voltage", 48)
        self.capacity_amp_hours = config.get("capacity_amp_hours", 50)
        self.num_batteries = config.get("num_batteries", 2)

    def add_power_event_listener(self, listener):
        """Add a power event listener.
        
        Args:
            listener (PowerEventListener): Listener to receive power event notifications
        """
        self.power_event_listeners.append(listener)

    def remove_power_event_listener(self, listener):
        """Remove a power event listener.
        
        Args:
            listener (PowerEventListener): Listener to remove
        """
        self.power_event_listeners.remove(listener)

    def notify_power_event_listeners(self, event: PowerEvent):
        """Notify all registered listeners of a power event.
        
        Args:
            event (PowerEvent): The power event to notify listeners about
        """
        for listener in self.power_event_listeners:
            listener.on_power_event(event)

    def is_estop_pressed(self):
        """Check if emergency stop is pressed based on battery voltage.
        
        Returns:
            bool: True if battery voltage is below ESTOP threshold, False otherwise
        """
        return self.last_state.last_reliable_seen_voltage < ESTOP_VOLTAGE_THRESHOLD

    def last_powered_up_time(self):
        """Get the last time the system was powered up.
        
        Returns the maximum of current boot time and last estop release time.
        
        Returns:
            float: Timestamp of last power-up event
        """
        return max(self.last_state.monitor_start_time, self.last_state.last_estop_released_time)
    
    def _step(self):
        """Update the last battery state."""
        prev_state, self.last_state = self.last_state, read_last_battery_state()
        if self.last_state.last_estop_pressed_time != prev_state.last_estop_pressed_time:
            info(
                f"E-stop pressed (voltage {self.last_state.last_reliable_seen_voltage:.3f} V at {time.time():.3f}s, "
                f"last_estop_pressed_time={self.last_state.last_estop_pressed_time})"
            )
            self.notify_power_event_listeners(PowerEvent.POWER_OFF)
        if self.last_state.last_estop_released_time != prev_state.last_estop_released_time:
            info(
                f"E-stop released (voltage {self.last_state.last_reliable_seen_voltage:.3f} V at {time.time():.3f}s, "
                f"last_estop_released_time={self.last_state.last_estop_released_time})"
            )
            self.notify_power_event_listeners(PowerEvent.POWER_RESTORED)
        if self.last_state.state_update_time - prev_state.state_update_time > 10.0:
            self.notify_power_event_listeners(PowerEvent.POWER_UNHEALTHY)
    
    def get_estimated_battery_percentage(self):
        """Get estimated battery percentage based on the last reliable seen voltage."""
        voltage = max(MIN_VOLTAGE, min(self.last_state.last_reliable_seen_voltage, MAX_VOLTAGE))
        return (voltage - MIN_VOLTAGE) / (MAX_VOLTAGE - MIN_VOLTAGE) * 100

class PowerMonitor():
    """Class used by feather deamon to monitor battery voltage"""

    def __init__(self):
        self.cfg = load_hydra_config("/feathersys/user_config/conf", "system/default.yaml")['system']['power']
        self.adc_type = self.cfg["adc_type"]
        self.hz = self.cfg["hz"]
        self.num_null_reads = 0
        self.last_state = BatteryState(
            monitor_start_time=time.time(),
            last_estop_pressed_time=0,
            last_estop_released_time=0,
            last_reliable_seen_voltage=-1,
            state_update_time=time.time(),
            error=None
        )

        if not POWER_STATE_PATH.parent.exists():
            POWER_STATE_PATH.parent.mkdir(parents=True, exist_ok=True)
            POWER_STATE_PATH.parent.chmod(0o777)
    
    def set_new_state(self, new_voltage: float, error: Optional[str] = None):
        if new_voltage < ESTOP_VOLTAGE_THRESHOLD and self.last_state.last_estop_pressed_time == 0:
            self.last_state.last_estop_pressed_time = time.time()  # If we start below estop threshold, trigger estop
        elif new_voltage < ESTOP_VOLTAGE_THRESHOLD and self.last_state.last_reliable_seen_voltage >= ESTOP_VOLTAGE_THRESHOLD:
            self.last_state.last_estop_pressed_time = time.time()  # Normal triggering of estop
        elif new_voltage > ESTOP_VOLTAGE_THRESHOLD and self.last_state.last_reliable_seen_voltage < ESTOP_VOLTAGE_THRESHOLD:
            self.last_state.last_estop_released_time = time.time()  # Normal releasing of estop
        
        self.last_state.error = error
        self.last_state.last_reliable_seen_voltage = new_voltage
        self.last_state.state_update_time = time.time()

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(json.dumps(asdict(self.last_state)))
            path = f.name
        os.rename(path, POWER_STATE_PATH)
        POWER_STATE_PATH.chmod(0o666)
    
    def step(self):
        new_voltage = get_battery_voltage(self.adc_type)

        if isinstance(new_voltage, str):
            self.set_new_state(0.0, error=new_voltage)
            return

        # If we read too many nulls in a row, set monitoring to unhealthy
        if new_voltage is None:
            self.num_null_reads += 1
            if self.num_null_reads > self.hz * 5:  # 5 seconds of null reads
                self.set_new_state(0.0)
                self.num_null_reads = 0
            return
        
        self.set_new_state(new_voltage)
        self.num_null_reads = 0



"""Virtual Robstride motor simulator."""
from feathersdk.robot.motors.robstride import RobstrideRunMode, RobstrideMotorMsg as MotorMsg, RobstrideMotorMode, \
    RobstrideMotorError, RobstrideVersionType
from feathersdk.robot.motors.robstride.params import RobstrideParamList, RobstrideFeedbackMessageBounds as FB_BOUNDS, \
    RobstrideOperationCommand as OP_CMD, get_param_info
from feathersdk.robot.motors.robstride.params import get_robstride_version
from feathersdk.robot.motors.robstride.message_handler import parse_robstride_can_id
from feathersdk.comms import CommsManager, SocketResult
from feathersdk.comms.system import is_can_interface, S_uint16be, S_uint16le
from dataclasses import dataclass
import math
from feathersdk.utils.common import try_log_exception_peacefully, currtime, timediff
from feathersdk.utils.feathertypes import *


DEFAULT_MOTOR_ARMATURE = 0.01  # kg⋅m²
MIN_DAMPING_FOR_STABILITY = 0.1  # Minimum damping to prevent instability
POSITION_ERROR_LOG_THRESHOLD = 10.0  # Log when error changes by this percentage


@dataclass
class VirtualRobstrideOperationState:
    """A dictionary of current operation mode values."""
    target_torque: float
    target_angle: float
    target_velocity: float
    kp: float
    kd: float

    def update(self, target_torque: float, target_angle: float, target_velocity: float, kp: float, kd: float) -> None:
        """Update the operation state with the given values."""
        self.target_torque = target_torque
        self.target_angle = target_angle
        self.target_velocity = target_velocity
        self.kp = kp
        self.kd = kd
    
    @staticmethod
    def default() -> "VirtualRobstrideOperationState":
        return VirtualRobstrideOperationState(
            target_torque=0.0,
            target_angle=0.0,
            target_velocity=0.0,
            kp=0.0,
            kd=0.0,
        )
    
    @staticmethod
    def from_op_command(op_cmd: OP_CMD) -> "VirtualRobstrideOperationState":
        """Create a VirtualRobstrideOperationState from a RobstrideOperationCommand."""
        return VirtualRobstrideOperationState(
            target_torque=op_cmd.target_torque,
            target_angle=op_cmd.target_angle,
            target_velocity=op_cmd.target_velocity,
            kp=op_cmd.kp,
            kd=op_cmd.kd,
        )


@dataclass
class VirtualRobstrideState:
    """A dictionary of current motor state, EG: current angle, velocity, torque, temperature, errors, etc."""
    enabled: bool
    angle: float
    velocity: float
    torque: float
    temp: float
    active_reporting: bool
    last_active_reporting_time: float
    motor_mode: RobstrideMotorMode
    errors: Set[RobstrideMotorError]
    params: RobstrideParamList

    def get_feedback_extra_byte(self) -> int:
        """Build the extra byte for the feedback message. First 2 bit for motor mode, then 6 bit for errors."""
        err_to_bit = {v: k for k, v in RobstrideMotorError._motor_error_bits().items()}
        return self.motor_mode.value << 6 | sum(err_to_bit.get(e, 0) for e in self.errors)
    
    def get_fault_feedback_bytes(self) -> bytes:
        """Get the fault feedback bytes for the fault feedback message.
        
        Note: I'm not exactly sure this is correct endianness since it's hard to trigger a fault manually, but based on
        some previous fault feedback messages we got once, it seems to be little endian.
        """
        err_to_bit = {v: k for k, v in RobstrideMotorError._fault_error_bits().items()}
        return S_uint16le.pack(sum(err_to_bit.get(e, 0) for e in self.errors)) + bytes([0])
    
    @staticmethod
    def default() -> "VirtualRobstrideState":
        return VirtualRobstrideState(
            enabled=False,
            angle=0.0,
            velocity=0.0,
            torque=0.0,
            temp=30.0,
            active_reporting=False,
            last_active_reporting_time=currtime(),
            motor_mode=RobstrideMotorMode.Reset,
            errors=set(),
            params=RobstrideParamList(),
        )
    
    def update(self, **kwargs) -> None:
        """Update the state with the given values."""
        for key, value in kwargs.items():
            if key not in self.__dataclass_fields__:
                raise ValueError(f"Invalid key: {key}")
            setattr(self, key, value)


def default_firmware_version(version: RobstrideVersionType) -> str:
    """Get the default firmware version a virtual motor should have for the given version."""
    version = get_robstride_version(version)
    return f"0.{version.value}.0.0"


def default_broadcast_id(version: RobstrideVersionType, motor_id: int) -> bytes:
    """Get the default broadcast ID for a virtual motor."""
    version = get_robstride_version(version)
    return bytes([0xDE, 0xAD, 0xBE, 0xEF, 0x00, version.value, 0x00, motor_id])


class VirtualRobstride:
    """Virtual Robstride motor simulator.
    
    Args:
        motor_id: The motor ID to simulate. Should be in range [0x00, 0xFF].
        version: The version of the Robstride motor to simulate.
        iface: The interface to use for the communication.

    Attributes:
        motor_id: The unique ID of this motor instance.
        version: The version of Robstride motor this instance is simulating.
        iface: The can interface to use for the communication.
        comms: The comms manager to use for the communication.
        params: A dictionary of current motor parameters and their values. Defaults to the param default values on
            initialization, with 'None' defaults overridden to a guessed default value.
        last_host_id: The last host ID we received a message from.
        state: A dictionary of current motor state, EG: current angle, velocity, torque, temperature, errors, etc.
    """
    def __init__(self, motor_id: int, version: RobstrideVersionType, iface: str):
        self.motor_id = motor_id
        self.version = get_robstride_version(version)
        self.firmware_version = default_firmware_version(self.version)
        self.iface = iface

        self.comms = CommsManager()
        self.comms.add_callback(self.handle_message)
        if not is_can_interface(iface):
            raise ValueError(f"Invalid interface: {iface}")
        self.comms.add_endpoint(iface, skip_duplicates=True)

        self.operation_vals = VirtualRobstrideOperationState.default()
        self.last_host_id = 0xAA

        self.state = VirtualRobstrideState.default()

        self._sent_first_broadcast_frame = False
        
        # Motor dynamics parameters
        self.motor_inertia = DEFAULT_MOTOR_ARMATURE  # kg⋅m² - motor inertia (affects acceleration)
        self.friction_coefficient = 0.5  # N⋅m⋅s/rad - viscous friction for damping
        
        # Logging control
        self._log_feedback_count = 0  # Counter to limit feedback logging
        
        # Position control error tracking
        self._last_error_percentage = None  # Track last error percentage for change detection
    
    def _vmsg_can_id(self, op: MotorMsg, motor_id: int, host_id: Optional[int] = None, e_byte: int = 0) -> int:
        """e_byte is the extra_byte value for the message."""
        host_id = self.last_host_id if host_id is None else host_id
        return (op.value << 24) | (e_byte << 16) | (motor_id << 8) | host_id

    def send_motor_feedback(self, msg_op: Optional[MotorMsg] = MotorMsg.MotorFeedback, send_ver: bool = False) -> None:
        """Send a motor feedback message. If `send_ver` is True, will send the firmware version number instead"""
        data = S_uint16be.map_and_pack(self.state.angle, *FB_BOUNDS.ANGLE_MAP[self.version]) + \
            S_uint16be.map_and_pack(self.state.velocity, *FB_BOUNDS.VELOCITY_MAP[self.version]) + \
            S_uint16be.map_and_pack(self.state.torque, *FB_BOUNDS.TORQUE_MAP[self.version]) + \
            S_uint16be.pack(int(self.state.temp * FB_BOUNDS.TEMP_DIVISOR))
        
        if send_ver:
            data = bytes([0, 0xC4, ord('V')] + list(map(int, self.firmware_version.split('.'))) + [7])
        
        can_id = self._vmsg_can_id(msg_op, self.motor_id, e_byte=self.state.get_feedback_extra_byte())
        self.comms.cansend(self.iface, True, can_id, data)
    
    def send_broadcast_frame(self) -> None:
        """Send a broadcast frame. Done after things like GetDeviceId and SetID."""
        can_id = self._vmsg_can_id(MotorMsg.GetDeviceId, self.motor_id, 0xFE)
        self.comms.cansend(self.iface, True, can_id, data=default_broadcast_id(self.version, self.motor_id))
    
    @try_log_exception_peacefully(capture_locals={'result', 'msg_op', 'motor_id', 'host_id', 'e_byte', 'param_info'})
    def handle_message(self, result: SocketResult):
        """Handle a message from the Robstride motor."""
        # Make sure this is a can message intended for this motor
        if not result.is_can() or not result.is_extended_can():
            return
    
        msg_op, motor_id, host_id, e_byte = parse_robstride_can_id(result.can_id, motor_to_host=False)
        if motor_id != self.motor_id:
            return

        self.last_host_id = host_id

        # Handle the given operation
        if msg_op == MotorMsg.GetDeviceId:
            self.send_broadcast_frame()

        elif msg_op == MotorMsg.OperationControl:
            # Note: these values can never be out of range because we use the entire 65535 range for each value
            self.operation_vals.update(
                target_torque=S_uint16be.unpack_and_map([e_byte, host_id], *FB_BOUNDS.TORQUE_MAP[self.version]),
                target_angle=S_uint16be.unpack_and_map(result.data[:2], *FB_BOUNDS.ANGLE_MAP[self.version]),
                target_velocity=S_uint16be.unpack_and_map(result.data[2:4], *FB_BOUNDS.VELOCITY_MAP[self.version]),
                kp=S_uint16be.unpack_and_map(result.data[4:6], *OP_CMD.KP_BOUNDS_MAP[self.version]),
                kd=S_uint16be.unpack_and_map(result.data[6:8], *OP_CMD.KD_BOUNDS_MAP[self.version]),
            )

        elif msg_op == MotorMsg.Enable:
            self.state.enabled = True
            self.state.motor_mode = RobstrideMotorMode.Run
            self.send_motor_feedback()
        
        elif msg_op == MotorMsg.Disable:
            self.state.enabled = False
            self.state.motor_mode = RobstrideMotorMode.Reset

            # Check if we should send the firmware version number back or not
            self.send_motor_feedback(send_ver=result.data[:2] == bytes([0, 0xC4]))
        
        elif msg_op == MotorMsg.ZeroPos:
            if result.data != bytes([0x01, 0, 0, 0, 0, 0, 0, 0]):
                raise ValueError(f"Invalid zero position command bytes: {result.data}")
            self.state.angle = 0.0
            self.send_motor_feedback()
        
        elif msg_op == MotorMsg.SetID:
            self.motor_id = e_byte
            self.send_broadcast_frame()
        
        elif msg_op == MotorMsg.ReadParam:
            p_info = get_param_info(S_uint16le.unpack(result.data[:2]))
            can_id = self._vmsg_can_id(MotorMsg.ReadParam, self.motor_id, host_id, e_byte=0)
            data = S_uint16le.pack(p_info.id) + bytes([0, 0]) \
                + p_info.dtype.pack(self.state.params[p_info.name].value, pad_end=4)
            self.comms.cansend(self.iface, True, can_id, data)
        
        elif msg_op == MotorMsg.WriteParam:
            p_info = get_param_info(S_uint16le.unpack(result.data[:2]))
            if not p_info.writable:
                raise ValueError(f"Param not writable: {p_info.name}")
            
            self.state.params[p_info.name].value = p_info.unpack_and_clamp(result.data[4:], self.version)
            self.send_motor_feedback()
        
        elif msg_op == MotorMsg.SaveMotorData:
            if result.data != bytes([1, 2, 3, 4, 5, 6, 7, 8]):
                raise ValueError(f"Invalid save motor data command bytes: {result.data}")
            print("Don't know what exactly happens with save motor data")
            self.send_motor_feedback()
        
        elif msg_op == MotorMsg.SetMotorBaudRate:
            if result.data[:-2] != bytes([1, 2, 3, 4, 5, 6]) or result.data[-2] not in [1, 2, 3, 4]:
                raise ValueError(f"Invalid set motor baud rate command bytes: {result.data}")
            print("Will ignore set motor baud rate command")
            self.send_broadcast_frame()
        
        elif msg_op == MotorMsg.SetMotorActiveReporting:
            if result.data[:-2] != bytes([1, 2, 3, 4, 5, 6]) or result.data[-2] not in [0, 1]:
                raise ValueError(f"Invalid set motor active reporting command bytes: {result.data}")
            self.send_motor_feedback(msg_op=MotorMsg.SetMotorActiveReporting)
            self.state.active_reporting = result.data[-2] == 1
            self.state.last_active_reporting_time = currtime()

        elif msg_op == MotorMsg.SetMotorProtocol:
            if result.data[:-2] != bytes([1, 2, 3, 4, 5, 6]) or result.data[-2] not in [0, 1, 2]:
                raise ValueError(f"Invalid set motor protocol command bytes: {result.data}")
            print("Will ignore set motor protocol command")
            self.send_broadcast_frame()
                
        else:
            pass
    
    def step(self, dt: float) -> None:
        """Step the motor simulation."""
        if not self._sent_first_broadcast_frame:
            self.send_broadcast_frame()
            self._sent_first_broadcast_frame = True

        position_error, target_pos = None, None

        if self.state.enabled:
            if self.state.params.run_mode.value == RobstrideRunMode.Position:
                position_error, target_pos = self._simulate_position_mode(dt)
            elif self.state.params.run_mode.value == RobstrideRunMode.Speed:
                position_error, target_pos = self._simulate_speed_mode(dt)
            elif self.state.params.run_mode.value == RobstrideRunMode.Operation:
                position_error, target_pos = self._simulate_operation_mode(dt)
            else:
                raise NotImplementedError(f"Unimplemented run mode: {self.state.params.run_mode}")
        
        # Log position error if applicable
        if position_error is not None and target_pos is not None:
            self._log_position_error(position_error, target_pos)
        
        # Send active reporting message if applicable
        st = self.state.params.epscan_time.value
        at_interval_secs = 0 if st == 0 else (st + 1) * 0.005
        
        if self.state.active_reporting and timediff(self.state.last_active_reporting_time) > at_interval_secs:
            self.send_motor_feedback(msg_op=MotorMsg.SetMotorActiveReporting)
            self.state.last_active_reporting_time = currtime()
        
        # Check for and send fault feedback message if applicable
        self._check_send_fault_feedback()
    
    def _check_send_fault_feedback(self) -> None:
        """Check if we should send a fault feedback message."""
        if len(self.state.errors) == 0:
            return
        
        can_id = self._vmsg_can_id(MotorMsg.FaultFeedback, self.motor_id, self.last_host_id, e_byte=0)
        self.comms.cansend(self.iface, True, can_id, self.state.get_fault_feedback_bytes() + bytes([0, 0, 0, 0, 0]))

        self.state.errors.clear()

    def _update_dynamics(self, dt: float) -> None:
        """Update motor dynamics based on current torque."""
        acceleration = self.state.torque / self.motor_inertia
        self.state.velocity = FB_BOUNDS.VELOCITY_MAP[self.version].clamp(self.state.velocity + acceleration * dt)
        self.state.angle = FB_BOUNDS.ANGLE_MAP[self.version].clamp(self.state.angle + self.state.velocity * dt)

    def _simulate_position_mode(self, dt: float) -> Tuple[Optional[float], Optional[float]]:
        """Simulate position control mode. Returns (position_error, target_pos)."""
        position_error = self.state.params.loc_ref.value - self.state.angle
        
        # PD control: desired velocity based on position error
        desired_velocity = position_error * self.state.params.loc_kp.value
        velocity_error = desired_velocity - self.state.velocity
        
        # Control torque from velocity error (using Kd as velocity gain), and friction damping
        # TODO: Figure out what this `kd` val is supposed to be, because I don't see it in position mode params
        control_torque = velocity_error * max(self.operation_vals.kd, MIN_DAMPING_FOR_STABILITY)
        friction_torque = -self.friction_coefficient * self.state.velocity
        
        # Total torque
        self.state.torque = FB_BOUNDS.TORQUE_MAP[self.version].clamp(control_torque + friction_torque)
        self._update_dynamics(dt)
        
        return position_error, self.state.params.loc_ref.value

    def _simulate_speed_mode(self, dt: float) -> Tuple[Optional[float], Optional[float]]:
        """Simulate speed control mode. Returns (position_error, target_pos)."""
        self.state.velocity = self.state.params.spd_ref.value
        self.state.angle = FB_BOUNDS.ANGLE_MAP[self.version].clamp(self.state.angle + self.state.velocity * dt)
        return None, None

    def _simulate_operation_mode(self, dt: float) -> Tuple[Optional[float], Optional[float]]:
        """Simulate operation control mode. Returns (position_error, target_pos)."""
        position_error = self.operation_vals.target_angle - self.state.angle
        vel_error = self.operation_vals.target_velocity - self.state.velocity
        
        # PD control torque and inherent friction damping (even when Kd=0, this provides stability)
        control_torque = position_error * self.operation_vals.kp + vel_error * self.operation_vals.kd
        friction_torque = -self.friction_coefficient * self.state.velocity
        
        # Total torque: feedforward + control + friction
        new_torque = self.operation_vals.target_torque + control_torque + friction_torque
        self.state.torque = FB_BOUNDS.TORQUE_MAP[self.version].clamp(new_torque)
        self._update_dynamics(dt)
        
        return position_error, self.operation_vals.target_angle

    def _log_position_error(self, position_error: float, target_pos: float):
        """Log position error when it changes significantly."""
        position_range = 8.0 * math.pi
        error_abs = abs(position_error)
        error_percentage = (error_abs / position_range) * 100.0 if position_range > 0 else 0.0
        
        if self._last_error_percentage is not None:
            error_change = abs(error_percentage - self._last_error_percentage)
            if error_change >= POSITION_ERROR_LOG_THRESHOLD:
                print(f"[MOTOR {self.motor_id:02X}]: Position error: {error_percentage:.2f}% "
                      f"(error={position_error:.4f} rad, target={target_pos:.4f} rad, "
                      f"actual={self.state.angle:.4f} rad, change={error_change:.2f}%)")
                self._last_error_percentage = error_percentage
        else:
            # First time, just record the value
            self._last_error_percentage = error_percentage



import logging
import struct
import time
from typing import Optional
from queue import Queue, Empty
from ....utils.files import write_json_file, read_json_file
from ....comms import CommsManager, SocketResult
from ....utils import constants
from .utils import (
    FeedbackFrameInfo,
    MotorErrorCode,
    MotorType,
    MotorConstants,
    uint_to_float,
    float_to_uint,
    ReceiveMode,
)
import os
import time
from feathersdk.utils.common import currtime
from feathersdk.utils.logger import warning, error, info
from feathersdk.utils.constants import MOTORS_CALIBRATION_PATH
import math
from feathersdk.robot.motors.motor import Motor



class ControlMode:
    MIT = "MIT"
    POS_VEL = "POS_VEL"
    VEL = "VEL"

    @classmethod
    def get_id_offset(cls, control_mode: str) -> int:
        if control_mode == cls.MIT:
            return 0x000
        elif control_mode == cls.POS_VEL:
            return 0x100
        elif control_mode == cls.VEL:
            return 0x200
        else:
            raise ValueError(f"Control mode '{control_mode}' not recognized.")


class DamiaoMotor(Motor):
    """Motor firmware interface for Damiao motors.
    
    Handles all low-level CAN communication with the motor firmware.
    """
    
    def __init__(
        self,
        motor_name: str,
        motor_id: int = 7,
        config: dict = {},
        iface: str = "can0",
    ):
        """Initialize the DamiaoMotor.

        Args:
            motor_name: Name of the motor (for calibration storage).
            motor_id: The CAN ID of the motor.
            motor_type: The type of the motor (MotorType.DM4310).
            iface: The CAN interface name (e.g., "can0").
            control_mode: The control mode (ControlMode.MIT, ControlMode.POS_VEL, ControlMode.VEL).
            receive_mode: The receive mode for response IDs (ReceiveMode.p16, etc.).
        """

        self.motor_name = motor_name
        self.motor_id = config.get("id", 7)
        self.iface = iface
        self.control_mode = ControlMode.MIT #so far only MIT is supported
        #use the version from the config
        self.version = config.get("version", "DM4310")
        if self.version == "DM4310":
            self.motor_type = MotorType.DM4310
        else:
            raise ValueError(f"Motor type '{self.version}' not recognized.")
        self.receive_mode = ReceiveMode.p16
        self.cmd_idoffset = ControlMode.get_id_offset(self.control_mode)

        self._homing_pos = config.get("homing_pos", 0) #i radians
        self.range = config.get("range", {
            "min": -math.pi / 2,
            "max": math.pi / 2,
        })

        self.lower_limit = None
        self.upper_limit = None
        self.total_range = None
        self.calibration_homing_pos = None
        self.calibration_time = None

        # Load motor constants from config, with defaults from MotorType
        self.motor_constants = self._get_motor_constants_from_config(config)

        self.dual_encoder: bool = self._get_dual_encoder_property()
        
        # Use provided CommsManager or create a new one
        self.comms = CommsManager()
        
        # Response queue for matching sent messages with received responses
        self._response_queue: Queue[SocketResult] = Queue()

        self.comms.add_callback(self._on_can_message)

        self.load_calibration_state()

    def _get_motor_constants_from_config(self, config: dict) -> MotorConstants:
        """Get MotorConstants from config, with defaults from MotorType.
        
        Args:
            config: Motor configuration dictionary.
            
        Returns:
            MotorConstants: Motor constants with config values if provided, defaults otherwise.
        """
        # Get default constants
        default_const = MotorType.get_motor_constants()
        
        # Get max values from config, fallback to defaults
        max_velocity = config.get("max_velocity", default_const.VELOCITY_MAX)
        max_torque = config.get("max_torque", default_const.TORQUE_MAX)
        
        # Min values are negative of max
        min_velocity = -max_velocity
        min_torque = -max_torque
        
        # Override with config values if provided
        return MotorConstants(
            POSITION_MAX=default_const.POSITION_MAX,
            POSITION_MIN=default_const.POSITION_MIN,
            VELOCITY_MAX=max_velocity,
            VELOCITY_MIN=min_velocity,
            TORQUE_MAX=max_torque,
            TORQUE_MIN=min_torque,
            KP_MAX=config.get("max_kp", default_const.KP_MAX),
            KP_MIN=default_const.KP_MIN,
            KD_MAX=config.get("max_kd", default_const.KD_MAX),
            KD_MIN=default_const.KD_MIN,
        )

    def _get_dual_encoder_property(self) -> bool:
        if self.motor_type == MotorType.DM4310:
            return True
        else:
            return False

    def _on_can_message(self, result: SocketResult) -> None:
        """Callback for receiving CAN messages from CommsManager.
        
        Filters messages for this motor and adds them to the response queue.
        """
        if result.is_error():
            return
        
        # Only process messages from our CAN interface
        if result.socket_name != self.iface:
            return
        
        # Check if this message is a response for our motor
        expected_receive_id = self.receive_mode.get_receive_id(self.motor_id)
        if result.can_id == expected_receive_id:
            self._response_queue.put(result)

    def _get_frame_id(self) -> int:
        """Calculate the Control Frame ID for this motor."""
        return self.cmd_idoffset + self.motor_id

    def _send_message_get_response(
        self,
        frame_id: int,
        data: bytes,
        max_retry: int = 5,
        expected_id: Optional[int] = None,
        timeout: float = 0.2,
    ) -> SocketResult:
        """Send a CAN message and wait for a response.

        Args:
            frame_id: The CAN frame ID to send.
            data: The data payload (8 bytes).
            max_retry: Maximum number of retry attempts.
            expected_id: Expected response CAN ID. If None, uses receive_mode.
            timeout: Timeout for each attempt in seconds.

        Returns:
            SocketResult: The received response message.

        Raises:
            AssertionError: If no response is received after max_retry attempts.
        """
        if expected_id is None:
            expected_id = self.receive_mode.get_receive_id(self.motor_id)
        
        # Ensure data is exactly 8 bytes
        if len(data) != 8:
            data = data[:8] + bytes(8 - len(data))
        
        for attempt in range(max_retry):
            try:
                # Clear any old messages from the queue
                while not self._response_queue.empty():
                    try:
                        self._response_queue.get_nowait()
                    except Empty:
                        break
                
                # Send the message
                self.comms.cansend(
                    self.iface,
                    extended=False,
                    can_id=frame_id,
                    data=data,
                )
                
                # Wait for response
                try:
                    response = self._response_queue.get(timeout=timeout)
                    if response.can_id == expected_id:
                        return response
                except Empty:
                    pass
                
                # Small delay before retry
                time.sleep(0.001)
                
            except Exception as e:
                warning(f"CAN send error (attempt {attempt + 1}/{max_retry}): {e}")
                time.sleep(0.001)
        
        raise AssertionError(
            f"Failed to communicate with motor {self.motor_id} on {self.iface} "
            f"after {max_retry} attempts"
        )

    def parse_recv_message(
        self,
        result: SocketResult,
        ignore_error: bool = False,
    ) -> FeedbackFrameInfo:
        """Parse the received message to extract motor information.

        Args:
            result: The SocketResult from CommsManager.
            ignore_error: If True, don't raise on error codes.

        Returns:
            FeedbackFrameInfo: The current state of the motor.
        """
        data = result.data
        if len(data) < 8:
            raise ValueError(f"Received message has insufficient data: {len(data)} bytes")
        
        error_int = (data[0] & 0xF0) >> 4
        error_hex = hex(error_int)
        error_message = MotorErrorCode.get_error_message(error_int)

        motor_id_of_this_response = self.receive_mode.to_motor_id(result.can_id)
        
        if error_hex != "0x1":
            warning(
                f"Motor id: {motor_id_of_this_response}, error: {error_message} "
                f"on {self.iface}"
            )
            if not ignore_error:
                error(
                    f"Motor id: {motor_id_of_this_response}, error: {error_message} "
                    f"on {self.iface}"
                )
                raise RuntimeError(
                    f"Motor error detected: motor name: {self.motor_name}, motor id: {motor_id_of_this_response}, "
                    f"error: {error_message}"
                )
        
        p_int = (data[1] << 8) | data[2]
        v_int = (data[3] << 4) | (data[4] >> 4)
        t_int = ((data[4] & 0xF) << 8) | data[5]
        temperature_mos = data[6]
        temperature_rotor = data[7]

        position = uint_to_float(p_int, self.motor_constants.POSITION_MIN, self.motor_constants.POSITION_MAX, 16)
        velocity = uint_to_float(v_int, self.motor_constants.VELOCITY_MIN, self.motor_constants.VELOCITY_MAX, 12)
        torque = uint_to_float(t_int, self.motor_constants.TORQUE_MIN, self.motor_constants.TORQUE_MAX, 12)
        temperature_mos = float(temperature_mos)
        temperature_rotor = float(temperature_rotor)

        return FeedbackFrameInfo(
            id=motor_id_of_this_response,
            error_code=error_hex,
            error_message=error_message,
            position=position,
            velocity=velocity,
            torque=torque,
            temperature_mos=temperature_mos,
            temperature_rotor=temperature_rotor,
        )

    def motor_on(self) -> FeedbackFrameInfo:
        """Turn on the motor.

        Returns:
            FeedbackFrameInfo: The motor state after enabling.
        """
        current_level = logging.getLogger().getEffectiveLevel()
        logging.getLogger().setLevel(logging.ERROR)

        frame_id = self.motor_id  # For motor_on, use motor_id directly
        data = bytes([0xFF] * 7 + [0xFC])

        try:
            response = self._send_message_get_response(frame_id, data)
            motor_info = self.parse_recv_message(response, ignore_error=True)
            
            if int(motor_info.error_code, 16) != MotorErrorCode.normal:
                while int(motor_info.error_code, 16) != MotorErrorCode.normal:
                    info(f"Motor {self.motor_id} error: {motor_info.error_message}")
                    self.clear_error()
                    time.sleep(0.01)
                    info(f"Motor {self.motor_id} error cleaned")
                    
                    response = self._send_message_get_response(frame_id, data)
                    motor_info = self.parse_recv_message(response, ignore_error=True)
            else:
                info(f"Motor {self.motor_id} is already on")
            
            logging.getLogger().setLevel(current_level)
            motor_info = self.parse_recv_message(response)
            return motor_info
        except Exception as e:
            logging.getLogger().setLevel(current_level)
            raise

    def clear_error(self) -> None:
        """Clear motor errors."""
        frame_id = self.motor_id
        data = bytes([0xFF] * 7 + [0xFB])
        info("Clearing error")
        
        for _ in range(3):
            try:
                self.comms.cansend(
                    self.iface,
                    extended=False,
                    can_id=frame_id,
                    data=data,
                )
            except Exception as e:
                warning(f"Error clearing motor error: {e}")

    def motor_off(self) -> None:
        """Turn off the motor."""
        frame_id = self._get_frame_id()
        data = bytes([0xFF] * 7 + [0xFD])
        self._send_message_get_response(frame_id, data)

    def save_zero_position(self) -> None:
        """Save the current position as zero position."""
        frame_id = self._get_frame_id()
        data = bytes([0xFF] * 7 + [0xFE])
        
        try:
            self._send_message_get_response(frame_id, data, timeout=2.0)
        except AssertionError:
            pass
        
        # Check if set zero position success
        current_state = self.set_control(pos=0, vel=0, kp=0, kd=0, torque=0)
        diff = abs(current_state.position)
        if diff < 0.01:
            info(
                f"Motor {self.motor_id} set zero position success, current position: {current_state.position}"
            )

    def set_control(
        self,
        pos: float = 0.0,
        vel: float = 0.0,
        kp: float = 0.0,
        kd: float = 0.0,
        torque: float = 0.0,
    ) -> FeedbackFrameInfo:
        """Set the control of the motor and return its status.

        Args:
            pos: The target position value.
            vel: The target velocity value.
            kp: The proportional gain value.
            kd: The derivative gain value.
            torque: The target torque value.

        Returns:
            FeedbackFrameInfo: The current state of the motor.
        """
        frame_id = self._get_frame_id()

        # Prepare the CAN message
        data = bytearray(8)
        if self.control_mode == ControlMode.MIT:
            pos_tmp = float_to_uint(pos, self.motor_constants.POSITION_MIN, self.motor_constants.POSITION_MAX, 16)
            vel_tmp = float_to_uint(vel, self.motor_constants.VELOCITY_MIN, self.motor_constants.VELOCITY_MAX, 12)
            kp_tmp = float_to_uint(kp, self.motor_constants.KP_MIN, self.motor_constants.KP_MAX, 12)
            kd_tmp = float_to_uint(kd, self.motor_constants.KD_MIN, self.motor_constants.KD_MAX, 12)
            tor_tmp = float_to_uint(torque, self.motor_constants.TORQUE_MIN, self.motor_constants.TORQUE_MAX, 12)
            data[0] = (pos_tmp >> 8) & 0xFF
            data[1] = pos_tmp & 0xFF
            data[2] = (vel_tmp >> 4) & 0xFF
            data[3] = ((vel_tmp & 0xF) << 4) | (kp_tmp >> 8)
            data[4] = kp_tmp & 0xFF
            data[5] = (kd_tmp >> 4) & 0xFF
            data[6] = ((kd_tmp & 0xF) << 4) | (tor_tmp >> 8)
            data[7] = tor_tmp & 0xFF

        # Send the CAN message and get response
        response = self._send_message_get_response(frame_id, bytes(data), max_retry=15)

        # Parse the received message
        motor_info = self.parse_recv_message(response)
        return motor_info

    def get_state(self) -> FeedbackFrameInfo:
        """Get the current motor state without releasing position hold.
        
        This method uses motor_on() to get feedback from the motor. Unlike calling
        set_control() with zeros, this approach doesn't release any active position
        hold that may be in effect. The motor should already be enabled, so this
        primarily serves to retrieve the current state (position, velocity, torque, 
        temperatures, etc.) without interfering with ongoing control commands.
        
        Returns:
            FeedbackFrameInfo: The current state of the motor.
        """
        # Use motor_on() to get feedback - this doesn't release position hold
        # like set_control() with zeros would
        return self.motor_on()

    def save_calibration_state(self, lower_limit: float, upper_limit: float, calibration_homing_pos: float, total_range: float):
        """Set the calibration data for this motor.
        
        Args:
            lower_limit: The encoder value at the lower physical limit
            upper_limit: The encoder value at the upper physical limit
            calibration_homing_pos: The encoder value at the homing position after calibration.
                This may have a small residual offset from 0 due to motor drift after zeroing.
            total_range: The total range of motion in radians
        """
        self._disable_compliance_mode = False
        self.calibration_time = currtime()
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.calibration_homing_pos = calibration_homing_pos
        self.total_range = total_range

        if not self.dual_encoder:
            warning(f"Motor {self.motor_name} does not have a dual encoder, skipping persistence of calibration data to disk.")
            return

        data = {
            "calibration_time": self.calibration_time,
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit,
            "calibration_homing_pos": self.calibration_homing_pos,
            "total_range": self.total_range,
        }
        write_json_file(data, MOTORS_CALIBRATION_PATH + "/" + self.motor_name + ".json")

    def load_calibration_state(self) -> bool:
        """Load the calibration state from the file.
        
        Returns True if the calibration state was loaded successfully, False otherwise.
        Returns False if the file doesn't exist or uses the old calibration format
        (middle_pos instead of calibration_homing_pos), requiring recalibration.
        """
        if not self.dual_encoder:
            return False

        file_path = MOTORS_CALIBRATION_PATH + "/" + self.motor_name + ".json"
        if not os.path.exists(file_path):
            return False
        data = read_json_file(file_path)
        # Check for new calibration format - return False if using old format
        if "calibration_homing_pos" not in data:
            return False
        self.calibration_time = data["calibration_time"]
        self.lower_limit = data["lower_limit"]
        self.upper_limit = data["upper_limit"]
        self.calibration_homing_pos = data["calibration_homing_pos"]
        self.total_range = data["total_range"]
        return True

    def check_in_range(self) -> bool:
        """Check if the motor is in range."""
        cur_pos = self.get_state().position
        if cur_pos is not None and self.lower_limit is not None and self.upper_limit is not None:
            if cur_pos >= self.lower_limit and cur_pos <= self.upper_limit:
                return True
        return False


import numpy as np
from scipy.interpolate import CubicSpline
from threading import Lock

def change_in_pos(v0, a0, jerk, t1, t2):
    """
    jerk = 0 if constant acceleration.
    """
    return v0 * (t2 - t1) + a0 * (t2 - t1)**2 / 2 + jerk * (t2 - t1)**3 / 6

def change_in_vel(a0, jerk, t1, t2):
    return a0 * (t2 - t1) + jerk * (t2 - t1)**2 / 2

def create_cubic_spline(min_value, max_value, num_steps):
    x_points = [0, 1]
    
    y_points = [min_value, max_value]
    cs = CubicSpline(x_points, y_points, bc_type='clamped')
    x_steps = np.linspace(0, 1, num_steps)
    y_steps = cs(x_steps)
    return y_steps

def clamp(x, lo, hi):
    return lo if x < lo else hi if x > hi else x

def create_cubic_hermite(p0, v0, p1, v1, num_steps):
    """
    Cubic Hermite interpolation from p0->p1 with endpoint slopes v0, v1
    over num_steps samples (including endpoints).
    v0 and v1 are in "value per step" units.
    """
    if num_steps <= 1:
        return [p1]

    traj = []
    n = num_steps - 1  # steps between endpoints
    for i in range(num_steps):
        t = i / n  # 0..1
        t2 = t * t
        t3 = t2 * t

        h00 = 2 * t3 - 3 * t2 + 1
        h10 = t3 - 2 * t2 + t
        h01 = -2 * t3 + 3 * t2
        h11 = t3 - t2

        # Scale slopes by total duration (n) to match Hermite formulation.
        # Here v0, v1 are per-step, so multiply by n.
        pt = (h00 * p0) + (h10 * (v0 * n)) + (h01 * p1) + (h11 * (v1 * n))
        traj.append(pt)
    return traj

class Trajectory:
    """
    A trajectory is a list of values that are used to control a motor.
    It is used to smoothly transition the motor from one position to another.

    Trajectories is based around calibrated_angle because we want to make it readable around the URDF.
    """

    def __init__(self, motor):
        self.index = 0
        self.motor = motor
        self.trajectory = []
        self.trajectory_lock = Lock()

        # Tunables
        self.max_slope = None       # e.g. set to a float to cap velocity changes per step
        self.end_slope = 0.0        # slope at the end (0 gives a gentle settle)

    def _current_value(self):
        # Prefer in-flight trajectory point if we have one, otherwise motor target.
        if self.trajectory and self.index < len(self.trajectory):
            return self.trajectory[self.index]
        return self.motor.convert_angle_to_calibrated_angle(self.motor.target_position)

    def _estimate_current_slope(self):
        """
        Estimate per-step slope from the existing trajectory near the current index.
        This gives continuity of velocity when retargeting mid-move.
        """
        if not self.trajectory:
            return 0.0

        i = self.index
        # Use a central/forward difference when possible
        if 1 <= i < len(self.trajectory) - 1:
            v = 0.5 * (self.trajectory[i + 1] - self.trajectory[i - 1])
        elif 0 <= i < len(self.trajectory) - 1:
            v = self.trajectory[i + 1] - self.trajectory[i]
        elif len(self.trajectory) >= 2:
            v = self.trajectory[-1] - self.trajectory[-2]
        else:
            v = 0.0

        if self.max_slope is not None:
            v = clamp(v, -self.max_slope, self.max_slope)
        return v

    def update_trajectory(self, end_value, num_steps):
        """Update the trajectory from the current position to go to the end_value over num_steps.

        Args:
            end_value: The target calibrated_angle to reach.
            num_steps: The number of steps to take to reach the end_value.
        """
        with self.trajectory_lock:
            p0 = self._current_value()
            v0 = self._estimate_current_slope()

            # Optional: also cap v1 (end slope)
            v1 = self.end_slope
            if self.max_slope is not None:
                v1 = clamp(v1, -self.max_slope, self.max_slope)

            self.index = 0
            self.trajectory = create_cubic_hermite(p0, v0, end_value, v1, num_steps)

    def step(self):
        with self.trajectory_lock:
            if len(self.trajectory) == 0:
                return self.motor.calibrated_angle[0]
            elif len(self.trajectory) >= self.index + 1:
                to_return = self.trajectory[self.index]
                self.index += 1
                return to_return
            else:
                return self.trajectory[-1]

    def reset(self):
        with self.trajectory_lock:
            self.index = 0
            self.trajectory = []

    def is_at_end(self):
        return self.index >= len(self.trajectory)

# class Trajectory:
#     """
#     A trajectory is a list of values that are used to control a motor.
#     It is used to smoothly transition the motor from one position to another.

#     Trajectories is based around calibrated_angle because we want to make it readable around the URDF.
#     """

#     def __init__(self, motor):
#         self.index = 0
#         self.motor = motor
#         self.trajectory = []
#         self.trajectory_lock = Lock()

#     def update_trajectory(self, end_value, num_steps):
#         """Update the trajectory from the current position to go to the end_value over num_steps.

#         Args:
#             end_value: The target calibrated_angle to reach.
#             num_steps: The number of steps to take to reach the end_value.
#         """
#         with self.trajectory_lock:
#             self.index = 0
#             self.trajectory = create_cubic_spline(self.motor.convert_angle_to_calibrated_angle(self.motor.target_position), end_value, num_steps)

#     def add_trajectory(self, end_value, num_steps):
#         """Add a trajectory to the current trajectory.

#         Args:
#             end_value: The target calibrated_angle to reach.
#             num_steps: The number of steps to take to reach the end_value.
#         """
#         with self.trajectory_lock:
#             self.trajectory = np.concatenate([self.trajectory, create_cubic_spline(self.trajectory[-1], end_value, num_steps)])

#     def step(self):
#         with self.trajectory_lock:
#             if len(self.trajectory) == 0:
#                 return self.motor.calibrated_angle[0]
#             elif len(self.trajectory) >= self.index + 1:
#                 to_return = self.trajectory[self.index]
#                 self.index += 1
#                 return to_return
#             else:
#                 return self.trajectory[-1]

#     def reset(self):
#         with self.trajectory_lock:
#             self.index = 0
#             self.trajectory = []

#     def is_at_end(self):
#         return self.index >= len(self.trajectory)


from machine import ADC

ADC_GP_PIN = 26  # pin 33, ADC0

# Linear map: divider_read at these points → battery voltage (V). Same slope extrapolates outside [48, 56].
_DIVIDER_LOW = 2.545
_DIVIDER_HIGH = 2.97
_VOLTAGE_LOW = 48.0
_VOLTAGE_HIGH = 56.0
_SLOPE = (_VOLTAGE_HIGH - _VOLTAGE_LOW) / (_DIVIDER_HIGH - _DIVIDER_LOW)

divider_read = ADC(ADC_GP_PIN).read_u16() * 3.3 / 65535
battery_voltage = _VOLTAGE_LOW + (divider_read - _DIVIDER_LOW) * _SLOPE
print(battery_voltage, divider_read)



from .steppable_system import SteppableSystem
from .motors.motors_manager import MotorsManager
from .motors.motor import Motor
from feathersdk.utils.feathertypes import Optional
from enum import Enum
import threading
import canopen
import os
import time
from feathersdk.utils import constants
from feathersdk.utils.common import currtime, timediff, dont_allow_on_fork, FeatherThread
from feathersdk.robot.battery_system import PowerEvent, BatterySystem
from feathersdk.utils.logger import info, error, warning
from feathersdk.robot.instance_lock import check_still_own_lock
from feathersdk.utils.feathertypes import *
import logging

MAX_POSITION_JUMP_THRESHOLD = 10_000
STEPS_PER_ROTATION = 65536
DEFAULT_RPS = 100  # RPS = rotations per second
DEFAULT_ACCELERATION_RPS2 = 50
DEFAULT_DECELERATION_RPS2 = 12

ROTATIONS_PER_CM_TRAVELED = 2.54  # 1 cm traveled per 2.54 rotation
MM_IN_CM = 10

ZERO_VELOCITY_THRESHOLD = 1000
ZERO_VELOCITY_MIN_SEQUENCE = 10

def overwrite_file(filepath: str, data: list):
    """Overwrite file with comma-separated data.
    
    Args:
        filepath (str): Path to the file to overwrite
        data (list): List of strings to write as comma-separated values
    """
    with open(filepath, "w") as file:
        file.write(",".join(data))


def read_array_from_file(filepath: str):
    """Read comma-separated values from file.
    
    Args:
        filepath (str): Path to the file to read
    
    Returns:
        list: List of strings read from the file
    """
    with open(filepath, "r") as file:
        return file.read().split(",")

class EZMotionControlMode(Enum):
    """EZMotion motor control modes."""
    POSITION = 1
    VELOCITY = 3
    HOMING = 6

class EZMotionCommands(Enum):
    """EZMotion CANopen command addresses."""
    SET_MAX_HOMING_TORQUE = 0x2070
    SET_OPERATION_STATE = 0x6040
    SET_CONTROL_MODE = 0x6060
    SET_POSITION = 0x607A
    SET_VELOCITY = 0x60FF
    SET_MAX_VELOCITY = 0x6081
    SET_MAX_ACCELERATION = 0x6083
    SET_MAX_DECELERATION = 0x6084
    GET_STATUS_WORD = 0x6041
    GET_CURRENT_POSITION = 0x6064
    GET_CURRENT_TORQUE = 0x6077
    GET_CURRENT_VELOCITY = 0x606C
    GET_CURRENT_CURRENT = 0x6078
    GET_TEMPERATURE = 0x2040
    GET_ERROR_REGISTER = 0x200B
    MAX_TORQUE = 0x6072
    HOMING_METHOD = 0x6098
    HOMING_SPEED = 0x6099
    HOMING_OFFSET = 0x607C
    HOMING_ACCELERATION = 0x609A

class EZMotionOperationState(Enum):
    """EZMotion motor operation states."""
    SHUTDOWN = 0x0006
    ENABLED = 0x000F
    TRANSITION_TO_NEW_POSITION = 0x001F

class EZMotionHomingMethod(Enum):
    """EZMotion homing methods."""
    HOME_USING_MAX_TORQUE_UP = -0x03
    HOME_USING_MAX_TORQUE_DOWN = -0x02


class EZMotionMotor(Motor):
    """EZMotion motor representation."""
    def __init__(self, motor_id: int, can_interface: str, motor_name: str, config: dict):
        """Initialize EZMotion motor.
        
        Args:
            motor_id (int): CAN node ID of the motor
            can_interface (str): CAN interface name
            motor_name (str): Name identifier for the motor
            config (dict): Motor configuration with optional keys:
                - default_target_velocity_rps: Default target velocity in rotations per second
                - default_target_acceleration_rps2: Default acceleration in rotations per second squared
                - default_target_deceleration_rps2: Default deceleration in rotations per second squared
        """
        self.motor_id = motor_id
        self.motor_name = motor_name
        self.target_velocity = config.get("default_target_velocity_rps", DEFAULT_RPS)
        self.target_acceleration = config.get("default_target_acceleration_rps2", DEFAULT_ACCELERATION_RPS2)
        self.target_deceleration = config.get("default_target_deceleration_rps2", DEFAULT_DECELERATION_RPS2)
        self.iface = can_interface

class CanOpenConditionalCommand:

    def __init__(self, parent, command: int, value: int, subindex: Optional[int] = None, skip_condition: callable = None, on_success: callable = None):
        self.parent = parent
        self.command = command
        self.value = value
        self.subindex = subindex
        self.success = False
        self.skip_condition = skip_condition
        self.on_success = on_success
        self.start_time = None
        self.end_time = None
    
    def mark_success(self):
        self.success = True
        self.end_time = currtime()
        if self.on_success is not None:
            self.on_success()

    def should_skip(self):
        if self.skip_condition is not None and self.skip_condition():
            self.mark_success()
            return True
        return False

    def execute(self):
        if self.start_time is None:
            self.start_time = currtime()
        if self.should_skip():
            return True
        try:
            if self.subindex is None:
                self.parent.node.sdo[self.command].raw = self.value
            else:
                self.parent.node.sdo[self.command][self.subindex].raw = self.value
            self.mark_success()
            info("command succeeded", self.command, self.value, self.subindex, self.end_time - self.start_time)
            return True
        except Exception as e:
            if self.parent.print_packet_loss:
                print(
                    "_safe_command() exception - ",
                    type(e).__name__ + ": " + str(e) + ".",
                    "Command: ",
                    self.command,
                    "Value: ",
                    self.value,
                )
            return False

class CanOpenTransaction:

    def __init__(self, conditional_commands: List[CanOpenConditionalCommand]):
        self.conditional_commands = conditional_commands
        self.successfully_completed = False
        self.start_time = None
        self.end_time = None

    def execute_next(self):
        # peek at the next command, execute it, and if it succeeds, pop it off the list and return True
        # if it fails, return False
        if self.start_time is None:
            self.start_time = currtime()
        if len(self.conditional_commands) == 0:
            return True
        command = self.conditional_commands[0]
        if command.execute():
            self.conditional_commands.pop(0)
            if len(self.conditional_commands) == 0:
                self.end_time = currtime()
                info("Transaction execution time: ", self.end_time - self.start_time)
                self.successfully_completed = True
            return True
        return False

    def execute(self):
        while self.execute_next():
            pass
        return self.successfully_completed

    def execute_next_n(self, n: int):
        for _ in range(n):
            self.execute_next()
        return self.successfully_completed


class TorsoPlatform(SteppableSystem):
    """Base class for torso platform control."""

    def __init__(self):
        """Initialize torso platform."""
        super().__init__()
        self.operating_frequency = 30

    def set_movement_profile(self, target_velocity, target_acceleration, target_jerk):
        pass

    def go_to(self, height_from_bottom: float):
        pass

    def recalibrate(self):
        pass

    async def get_position(self):
        pass

    async def get_state(self):
        pass

    async def health_check(self):
        pass

class EZMotionTorsoPlatform(TorsoPlatform):
    """EZMotion-based torso platform implementation."""

    def __init__(self, motors_manager: MotorsManager, power: BatterySystem, can_interface: Optional[str], config: dict):
        """Initialize EZMotion torso platform.
        
        Args:
            motors_manager (MotorsManager): Manager for motor communication
            power (BatterySystem): Battery system for power event handling
            can_interface (Optional[str]): CAN interface name, or None to skip CAN initialization
            config (dict): Configuration dictionary with required "motors" key and optional keys:
                - top_height_mm: Maximum height in millimeters (default: 600)
                - bottom_height_mm: Minimum height in millimeters (default: 50)
                - homing_offset_from_top_num_rotations: Homing offset from top in rotations (default: 1)
                - homing_max_torque_percent: Maximum torque for homing in percent (default: 1000)
                - homing_acceleration: Homing acceleration in rotations per second squared (default: 50)
                - print_packet_loss: If True, print packet loss warnings (default: False)
                - show_all_sdo_errors: If True, show all SDO errors (default: True)
        
        Raises:
            Exception: If power cycle detected and system must be recalibrated
        """
        self.cfg = config
        self.is_forked = False
        self.motors_map = {}
        self.power = power
        self.top_height_mm = self.cfg.get("top_height_mm", 600)
        self.bottom_height_mm = self.cfg.get("bottom_height_mm", 50)
        self.homing_offset_from_top_num_rotations = self.cfg.get("homing_offset_from_top_num_rotations", 1)
        self.homing_max_torque_percent = self.cfg.get("homing_max_torque_percent", 1000)
        self.homing_acceleration_rps2 = self.cfg.get("homing_acceleration", 50)
        self.print_packet_loss = self.cfg.get("print_packet_loss", False)
        self.last_sent_velocity = 0
        self.last_sent_acceleration = 0
        self.last_sent_deceleration = 0
        self.last_sent_position = 0

        for motor_name, motor_dict in self.cfg["motors"].items():
            self.main_motor_id = int(motor_dict["id"])
            self.main_motor_model = motor_dict["model"]
            self.motors_map[motor_name] = EZMotionMotor(self.main_motor_id, can_interface, motor_name, motor_dict["motor_config"])
        self.calibration_file_path = constants.MOTORS_CALIBRATION_PATH + "/_ezmotion_TvC.json"
        self.motors_manager = motors_manager
        self.motors_manager.add_motors(self.motors_map, family_name="torso")
        self.health_check_future = None
        self.zero_velocity_count = 0
        self.reference_position = None
        # self.comms = CommsManager()
        self.network = canopen.Network()

        if can_interface is not None:
            self.network.connect(channel=can_interface, bustype="socketcan", bitrate=1000000)
            eds_file = os.path.join(os.path.dirname(__file__), "motors", self.main_motor_model + ".eds")
            self.node = self.network.add_node(self.main_motor_id, eds_file)
        else:
            self.node = None
        
        self.healthy_state = False
        # assigned_velocity was the last velocity the user requested.
        self.assigned_velocity = 0
        # last_registered_velocity is the velocity that was confirmed received by the motor.
        self.last_registered_velocity = (0, -1)
        self.last_step_time = currtime()

        self.current_position = None
        try:
            last_state = read_array_from_file(self.calibration_file_path)
            if self.power is None:
                raise Exception("Error: Power system not initialized.")
            elif self.power.last_powered_up_time() > float(last_state[2]):
                raise Exception("Error: Power cycle detected. System must be recalibrated.")
            
            if last_state[0] == "recalibrated":
                self.current_position = self.get_current_position_with_retry()
                self.curr_height = float(last_state[1])
                self._set_reference_position(self.current_position, self.curr_height)
                #self.healthy_state = True
            if last_state[0] == "finished_moving":
                self.current_position = self.get_current_position_with_retry()
                self.curr_height = float(last_state[1])
                self._set_reference_position(self.current_position, self.curr_height)
                #self.healthy_state = True
        except Exception as e:
            error(f"Torso must be recalibrated: {str(e)}")

        self.operation_state = EZMotionOperationState.SHUTDOWN
        self.last_command = None
        self.init_ezmotion()
        self.moving = False
        self.control_mode = EZMotionControlMode.POSITION
        self.recalibrating = False
        self.aborted = False

        # Suppress all logging from canopen. TODO: Filter the messages to keep the ones we want to worry about.
        # No CRITICAL logging messages in canopen 2.4.1, so this should disable all sdo.client/server messages.
        sdo_level = logging.INFO if self.cfg.get("show_all_sdo_errors", True) else logging.CRITICAL
        logging.getLogger("canopen.sdo.client").setLevel(sdo_level)
        logging.getLogger("canopen.sdo.server").setLevel(sdo_level)

        self.height = (None, 0, 0, 0) # (height, timestamp, num_updates)
        self.velocity = (None, 0, 0, 0)
        self.torque = (None, 0, 0, 0)
        self.temperature = (None, 0, 0, 0)
        self.position = (None, 0, 0, 0)

        self.transaction_lock = threading.Lock()
        self.transaction = None
        self._canopen_stop = False
        self.canopen_thread = None

        super().__init__()

        if self.node is not None:
            self.canopen_thread = FeatherThread(target=self.canopen_thread_main, args=(self.node,))
            self.canopen_thread.start()


    @dont_allow_on_fork
    def canopen_thread_main(self, node: canopen.Node):
        try:
            next_clock_time = 0
            cnt = 0
            printed_at = -1
            while not self._canopen_stop:
                if cnt % 40 == 0 and printed_at != cnt:
                    printed_at = cnt
                if cnt % (self.operating_frequency / 20) == 0:
                    if not check_still_own_lock():
                        error("Torso: Lock not owned by this process. Skipping CANopen thread.")
                        break
                now = currtime()
                if now > next_clock_time:
                    try:
                        # For some reason, this has to be here to detect homing was successful, and we need to
                        # figure out why TODO:
                        status_word = self.get_status_word()
                        if self.transaction is not None:
                            self.transaction.execute_next_n(6)
                        position = self.get_current_position()
                        if position is not False:
                            self.position = (position, currtime(), self.position[2] + 1)
                            #if self.current_position is not None and \
                            #    abs(self.position[0] - self.current_position) > MAX_POSITION_JUMP_THRESHOLD and self.healthy_state:
                            #    error("Position jump detected, needs recalibration", self.position[0], self.current_position)
                            #     self.healthy_state = False
                            self.current_position = self.position[0]
                        if self.reference_position is not None:
                            height = self._compute_current_height(self.position[0])
                            if height is not False:
                                # Height's timestamp is the same as the position's timestamp.
                                self.height = (height, self.position[1], self.position[2])

                        velocity = self._safe_read(EZMotionCommands.GET_CURRENT_VELOCITY.value)
                        velocity_mm_per_second = velocity / 65536 / 2.54 * -10
                        # We ignore velocities greater than 200 mm/s to avoid extreme sensor noise coming from the motor.
                        if velocity is not False and abs(velocity_mm_per_second) < 200:
                            self.velocity = (velocity_mm_per_second, currtime(), self.velocity[2] + 1)
                        temp = self._safe_read(EZMotionCommands.GET_TEMPERATURE.value)
                        if temp is not False:
                            self.temperature = (temp, currtime(), self.temperature[2] + 1)
                        torque = self.get_current_torque()
                        if torque is not False:
                            self.torque = (torque, currtime(), self.torque[2] + 1)

                    except Exception as e:
                        error(f"Error in canopen thread: {e}")
                    next_clock_time = now + 1.0 / self.operating_frequency
                    cnt += 1
                time.sleep(0.001)
        except Exception as e:
            error(f"super-Error in canopen thread: {e}")
        finally:
            info("Torso: Exiting canopen_thread_main")
    
    @dont_allow_on_fork
    def _get_node(self):
        """Get CANopen node.
        
        Returns:
            canopen.Node: The CANopen node instance
        
        Raises:
            ValueError: If node is not initialized
        """
        if self.node is None:
            raise ValueError("Node is not initialized")
        return self.node

    @dont_allow_on_fork
    def init_ezmotion(self):
        """Initialize EZMotion motor to position control mode."""
        if self.node is None:
            error("Node is not initialized, skipping EZMotion initialization")
            return
        self.shutdown()
        self.set_control_mode(EZMotionControlMode.POSITION)
        # self.enable()
        # self.set_movement_profile("TvC", DEFAULT_RPS, DEFAULT_ACCELERATION_RPS2, DEFAULT_DECELERATION_RPS2)

    @dont_allow_on_fork
    def send_movement_profile(self, target_velocity=None, target_acceleration=None, target_deceleration=None):
        """Send movement profile parameters to motor (only if changed).
        
        Args:
            target_velocity (float, optional): Target velocity in rotations per second
            target_acceleration (float, optional): Target acceleration in rotations per second squared
            target_deceleration (float, optional): Target deceleration in rotations per second squared
        """
        if target_velocity is not None:
            if target_velocity != self.last_sent_velocity:
                if self._ensure_command(EZMotionCommands.SET_MAX_VELOCITY.value, target_velocity * STEPS_PER_ROTATION): 
                    self.last_sent_velocity = target_velocity
        if target_acceleration is not None:
            if target_acceleration != self.last_sent_acceleration:
                if self._ensure_command(EZMotionCommands.SET_MAX_ACCELERATION.value, target_acceleration * STEPS_PER_ROTATION):
                    self.last_sent_acceleration = target_acceleration
        if target_deceleration is not None:
            if target_deceleration != self.last_sent_deceleration:
                if self._ensure_command(EZMotionCommands.SET_MAX_DECELERATION.value, target_deceleration * STEPS_PER_ROTATION):
                    self.last_sent_deceleration = target_deceleration

    @dont_allow_on_fork
    def set_movement_profile(self, target_velocity, target_acceleration, target_deceleration):
        """Set movement profile parameters for the motor.
        
        Args:
            target_velocity (float, optional): Target velocity in rotations per second
            target_acceleration (float, optional): Target acceleration in rotations per second squared
            target_deceleration (float, optional): Target deceleration in rotations per second squared
        """
        motor = self.motors_map["TvC"]
        if target_velocity is not None:
            motor.target_velocity = target_velocity
        if target_acceleration is not None:
            motor.target_acceleration = target_acceleration
        if target_deceleration is not None:
            motor.target_deceleration = target_deceleration

    @dont_allow_on_fork
    def _step(self):
        """Private method to monitor movement and handle target reached events."""
        try:
            self.last_step_time = currtime()
            if self.moving:
                status_word = self.get_status_word()
                if status_word & (1 << 10):
                    info("target reached")
                    if self.recalibrating:
                        self.shutdown()
                        self._on_recalibrated()
                    elif self.control_mode == EZMotionControlMode.POSITION:
                        self._on_target_reached()
                else:
                    # 3 reads
                    current_torque = self.torque[0] # if self.torque is not None else self.get_current_torque()
                    if current_torque > 1000 or current_torque < -1000:
                        info(f"calibration triggered {current_torque}")
                        self.shutdown()
                        if self.recalibrating:
                            self._on_recalibrated()
                    current_velocity = self.velocity[0] # if self.velocity is not None else self.get_current_velocity()
                    self.current_position = self.position[0] #if self.position is not None else self.get_current_position()
                    if current_velocity is not None and abs(current_velocity) < ZERO_VELOCITY_THRESHOLD:
                        self.zero_velocity_count += 1
                    else:
                        self.zero_velocity_count = 0
                    if self.zero_velocity_count > ZERO_VELOCITY_MIN_SEQUENCE and (self.last_command is not None and self.last_command[0] != "recalibrate"):
                        self.moving = False
                        self._on_target_reached()
        except Exception as e:
            import traceback
            error("Error in _step: ", traceback.format_exc())

    @dont_allow_on_fork
    def _ensure_command(self, command: int, value: int, subindex: Optional[int] = None, retries: int = 5):
        """Send command with retry logic.
        
        Args:
            command (int): CANopen command address
            value (int): Value to write
            subindex (Optional[int]): Subindex for the command
            retries (int): Number of retry attempts (default: 5)
        
        Returns:
            bool: True if command succeeded, False otherwise
        """
        for i in range(retries):
            if self._safe_command(command, value, subindex):
                return True
        return False

    @dont_allow_on_fork
    def _safe_command(self, command: int, value: int, subindex: Optional[int] = None):
        """Send command to motor with error handling.
        
        Args:
            command (int): CANopen command address
            value (int): Value to write
            subindex (Optional[int]): Subindex for the command
        
        Returns:
            bool: True if command succeeded, False otherwise
        """
        try:
            if subindex is None:
                self.node.sdo[command].raw = value
            else:
                self.node.sdo[command][subindex].raw = value
            return True
        except Exception as e:
            if self.print_packet_loss:
                print(
                    "_safe_command() exception - ",
                    type(e).__name__ + ": " + str(e) + ".",
                    "Command: ",
                    command,
                    "Value: ",
                    value,
                )
            return False

    @dont_allow_on_fork
    def _safe_read(self, command: int, subindex: Optional[int] = None):
        """Read value from motor with error handling.
        
        Args:
            command (int): CANopen command address to read from
        
        Returns:
            int | bool: Read value, or False if read failed
        """
        try:
            if subindex is None:
                return self._get_node().sdo[command].raw
            else:
                return self._get_node().sdo[command][subindex].raw
        except Exception as e:
            return False
    
    @dont_allow_on_fork
    def has_error(self):
        """Check if motor has error."""
        return self._safe_read(EZMotionCommands.GET_ERROR_REGISTER.value, 0x0B) != 0
    
    @dont_allow_on_fork
    def clear_errors(self):
        """Clear motor errors."""
        self._safe_command(EZMotionCommands.SET_OPERATION_STATE.value, 128)

    @dont_allow_on_fork
    def _is_position_control_mode(self):
        return self.control_mode == EZMotionControlMode.POSITION
    
    @dont_allow_on_fork
    def _is_enabled(self):
        return self.operation_state == EZMotionOperationState.ENABLED

    @dont_allow_on_fork
    def _record_control_mode(self, control_mode: EZMotionControlMode):
        self.control_mode = control_mode
    
    @dont_allow_on_fork
    def _record_operation_state(self, operation_state: EZMotionOperationState):
        self.operation_state = operation_state

    @dont_allow_on_fork
    def _record_velocity(self, velocity: float):
        self.last_sent_velocity = velocity

    @dont_allow_on_fork
    def _record_acceleration(self, acceleration: float):
        self.last_sent_acceleration = acceleration

    @dont_allow_on_fork
    def _record_deceleration(self, deceleration: float):
        self.last_sent_deceleration = deceleration

    @dont_allow_on_fork
    def _record_last_sent_position(self, position: int):
        self.last_sent_position = position
    
    @dont_allow_on_fork
    def _record_recalibrating(self, recalibrating: bool):
        self.recalibrating = recalibrating
        info(f"started record recalibrating: {recalibrating}")
        overwrite_file(self.calibration_file_path, ["recalibrating", str(0), str(time.time())])

    @dont_allow_on_fork
    def _record_moving(self, moving: bool):
        self.moving = moving

    @dont_allow_on_fork
    def _go_to_raw_position(self, position: int, velocity=None):
        """Move to raw position in motor steps.
        
        Args:
            position (int): Target position in motor steps
            velocity (float, optional): Velocity in rotations per second. If None, uses motor's default velocity.
        """
        self.commands = [
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_CONTROL_MODE.value, 
                EZMotionControlMode.POSITION.value, 
                skip_condition=self._is_position_control_mode,
                on_success=lambda: self._record_control_mode(EZMotionControlMode.POSITION)),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_OPERATION_STATE.value, 
                EZMotionOperationState.ENABLED.value, 
                skip_condition=self._is_enabled,
                on_success=lambda: self._record_operation_state(EZMotionOperationState.ENABLED)),
        ]
        if velocity is not None:
            self.commands.append(CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_MAX_VELOCITY.value, 
                velocity * STEPS_PER_ROTATION,
                skip_condition=lambda: velocity == self.last_sent_velocity,
                on_success=lambda: self._record_velocity(velocity)))
        else:
            motor = self.motors_map["TvC"]
            self.commands.append(CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_MAX_VELOCITY.value, 
                motor.target_velocity * STEPS_PER_ROTATION,
                skip_condition=lambda: motor.target_velocity == self.last_sent_velocity,
                on_success=lambda: self._record_velocity(motor.target_velocity)))
            self.commands.append(CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_MAX_ACCELERATION.value, 
                motor.target_acceleration * STEPS_PER_ROTATION,
                skip_condition=lambda: motor.target_acceleration == self.last_sent_acceleration,
                on_success=lambda: self._record_acceleration(motor.target_acceleration)))
            self.commands.append(CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_MAX_DECELERATION.value, 
                motor.target_deceleration * STEPS_PER_ROTATION,
                skip_condition=lambda: motor.target_deceleration == self.last_sent_deceleration,
                on_success=lambda: self._record_deceleration(motor.target_deceleration)))
       
        self.commands.append(CanOpenConditionalCommand(
            self,
            EZMotionCommands.SET_POSITION.value, 
            position,
            skip_condition=lambda: self.last_sent_position == position,
            on_success=lambda: self._record_last_sent_position(position)))

        self.commands.append(CanOpenConditionalCommand(
            self,
            EZMotionCommands.SET_OPERATION_STATE.value, 
            EZMotionOperationState.ENABLED.value,
            on_success=lambda: self._record_operation_state(EZMotionOperationState.ENABLED)))
        self.commands.append(CanOpenConditionalCommand(
            self,
            EZMotionCommands.SET_OPERATION_STATE.value, 
            EZMotionOperationState.TRANSITION_TO_NEW_POSITION.value,
            on_success=lambda: (self._record_operation_state(EZMotionOperationState.TRANSITION_TO_NEW_POSITION), self._record_moving(True))))
        self.transaction = CanOpenTransaction(self.commands)

    @dont_allow_on_fork
    def _compute_current_height(self, current_position: int = None):
        """Compute current height from reference position.
        
        Returns:
            float: Current height in millimeters
        """
        if current_position is None:
            self.current_position = self.get_current_position()
        num_rotations_offset = (
            self.current_position - self.reference_position["associated_position"]
        ) / STEPS_PER_ROTATION
        return (
            self.reference_position["associated_height"] - num_rotations_offset / ROTATIONS_PER_CM_TRAVELED * MM_IN_CM
        )

    @dont_allow_on_fork
    def _on_target_reached(self):
        """Handle target reached event and save state."""
        self.current_position = self.get_current_position_with_retry()
        num_rotations_offset = (
            self.current_position - self.reference_position["associated_position"]
        ) / STEPS_PER_ROTATION
        self.curr_height = (
            self.reference_position["associated_height"] - num_rotations_offset / ROTATIONS_PER_CM_TRAVELED * MM_IN_CM
        )
        overwrite_file(self.calibration_file_path, ["finished_moving", str(self.curr_height), str(time.time())])
        self.shutdown()

    @dont_allow_on_fork
    def _set_reference_position(self, current_position: int, current_height: float):
        """Set reference position for height calculations.
        
        Args:
            current_position (int): Current motor position in steps
            current_height (float): Current height in millimeters
        """
        self.reference_position = {"associated_position": current_position, "associated_height": current_height}
        self.max_height_target_position = self.calc_max_height_target_position(current_height, current_position)
        self.min_height_target_position = self.calc_min_height_target_position(current_height, current_position)
        info("max min target positions", self.max_height_target_position, self.min_height_target_position)

    @dont_allow_on_fork
    def _on_recalibrated(self):
        """Handle recalibration completion and save state."""
        info("on_recalibrated torso")
        self.current_position = self.get_current_position_with_retry()
        # self.min_position = current_position + (NUM_ROTATIONS_OFFSET_FROM_TOP * STEPS_PER_ROTATION)
        self.curr_height = self.top_height_mm
        self._set_reference_position(self.current_position, self.curr_height)
        overwrite_file(self.calibration_file_path, ["recalibrated", str(self.curr_height), str(time.time())])
        # Motor's target position should now be the current position.
        self.healthy_state = True
        self.recalibrating = False
        info("Torso calibration successful")

    @dont_allow_on_fork
    def wait_for_recalibration(self):
        """Wait for recalibration to complete."""
        while self.recalibrating and not self.aborted and not self.is_forked:
            time.sleep(0.1)

    @dont_allow_on_fork
    def recalibrate(self):
        """Recalibrate torso by homing to top position."""
        if self.node is None:
            raise ValueError("Node is not initialized; cannot recalibrate torso.")
        self.last_command = ["recalibrate", currtime()]
        # Mark recalibration as in-progress immediately so callers that invoke
        # wait_for_recalibration() right after this will actually block.
        self._record_recalibrating(True)
        commands = [
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_OPERATION_STATE.value,
                EZMotionOperationState.ENABLED.value,
                skip_condition=lambda: self.operation_state == EZMotionOperationState.ENABLED,
                on_success=lambda: self._record_operation_state(EZMotionOperationState.ENABLED)),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_OPERATION_STATE.value,
                EZMotionOperationState.SHUTDOWN.value,
                skip_condition=lambda: self.operation_state == EZMotionOperationState.SHUTDOWN,
                on_success=lambda: self._record_operation_state(EZMotionOperationState.SHUTDOWN)),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_CONTROL_MODE.value,
                EZMotionControlMode.HOMING.value,
                skip_condition=lambda: self.control_mode == EZMotionControlMode.HOMING,
                on_success=lambda: self._record_control_mode(EZMotionControlMode.HOMING)),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_OPERATION_STATE.value,
                EZMotionOperationState.ENABLED.value,
                skip_condition=lambda: self.operation_state == EZMotionOperationState.ENABLED,
                on_success=lambda: self._record_operation_state(EZMotionOperationState.ENABLED)),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.HOMING_METHOD.value,
                EZMotionHomingMethod.HOME_USING_MAX_TORQUE_UP.value),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_MAX_HOMING_TORQUE.value,
                self.homing_max_torque_percent,
                subindex=0x01),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.HOMING_SPEED.value,
                STEPS_PER_ROTATION,
                subindex=0x01),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.HOMING_SPEED.value,
                STEPS_PER_ROTATION,
                subindex=0x02),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.HOMING_ACCELERATION.value,
                self.homing_acceleration_rps2 * STEPS_PER_ROTATION),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.HOMING_OFFSET.value,
                self.homing_offset_from_top_num_rotations * STEPS_PER_ROTATION),
            CanOpenConditionalCommand(
                self,
                EZMotionCommands.SET_OPERATION_STATE.value,
                EZMotionOperationState.TRANSITION_TO_NEW_POSITION.value,
                on_success=lambda: self._record_moving(True)),
        ]
        self.transaction = CanOpenTransaction(commands)
        info("recalibrating torso")

    @dont_allow_on_fork
    def _go_to_raw_position_safely(self, position: int):
        """Move to raw position using safe low-speed profile.
        
        Args:
            position (int): Target position in motor steps
        """
        if self.control_mode != EZMotionControlMode.POSITION:
            self.set_control_mode(EZMotionControlMode.POSITION)
        self.enable()
        self._safe_command(EZMotionCommands.SET_POSITION.value, position)
        motor = self.motors_map["TvC"]
        self.send_movement_profile(5, 5, 5)
        self.transition_to_new_position()

    @dont_allow_on_fork
    def enable_velocity_mode(self):
        """Enable velocity control mode and set velocity to zero."""
        self.wait_for_recalibration()
        self.enable()
        if self._safe_command(EZMotionCommands.SET_VELOCITY.value, 0 * STEPS_PER_ROTATION):
            self.last_registered_velocity = (0, currtime())
        motor = self.motors_map["TvC"]
        self.send_movement_profile(motor.target_velocity, motor.target_acceleration, motor.target_deceleration)
        self.transition_to_new_position()

    @dont_allow_on_fork
    def _get_max_safe_height(self):
        """Get maximum safe height position accounting for spring compression.
        
        Returns:
            int: Maximum safe position in motor steps
        """
        return self.max_height_target_position + (STEPS_PER_ROTATION * 2)

    @dont_allow_on_fork
    def _get_min_safe_height(self):
        """Get minimum safe height position accounting for spring expansion.
        
        Returns:
            int: Minimum safe position in motor steps
        """
        return self.min_height_target_position - (STEPS_PER_ROTATION * 4)

    @dont_allow_on_fork
    def _is_at_max_height(self):
        """Check if at maximum safe height.
        
        Returns:
            bool: True if at or above maximum safe height
        """
        return self.current_position <= self._get_max_safe_height()

    @dont_allow_on_fork
    def _is_at_min_height(self):
        """Check if at minimum safe height.
        
        Returns:
            bool: True if at or below minimum safe height
        """
        return self.current_position >= self._get_min_safe_height()

    @dont_allow_on_fork
    def update_velocity(self, mm_per_second: float, skip_overwrite: bool = False):
        """Update torso velocity in velocity control mode.
        
        Args:
            mm_per_second (float): Velocity in millimeters per second, range -200 to 200
            skip_overwrite (bool): If True, skip overwriting calibration file
        
        Raises:
            Exception: If system not healthy or velocity exceeds maximum
        """
        if not self.healthy_state:
            raise Exception("Error: System must be recalibrated.")
        if mm_per_second > 200 or mm_per_second < -200:
            raise Exception("Error: Velocity is too high. Maximum velocity is 200 mm/s.")
        self.assigned_velocity = mm_per_second
        if self.current_position is not False:
            if (self._is_at_max_height() and mm_per_second > 0) or (self._is_at_min_height() and mm_per_second < 0):
                return
        rotations_per_second = mm_per_second / MM_IN_CM * ROTATIONS_PER_CM_TRAVELED
        # TODO: Currently, if we overflow the CAN line with too many commands, reading the position feedback will be delayed for the clips.
        target_velocity = -1 * rotations_per_second * STEPS_PER_ROTATION
        if self.last_command is None or self.last_command[0] != "update_velocity" or self.last_command[1] != target_velocity:
            self.last_command = ["update_velocity", target_velocity]
            if not skip_overwrite:
                overwrite_file(self.calibration_file_path, ["update_velocity", str(target_velocity), str(time.time())])
            if mm_per_second > 0:
                self._go_to_raw_position(
                    self.max_height_target_position, rotations_per_second
                )
            elif mm_per_second < 0:
                self._go_to_raw_position(
                    self.min_height_target_position, -1 * rotations_per_second
                )
            else:
                self._go_to_raw_position(self.current_position, 0)
    
    @dont_allow_on_fork
    def update_relative(self, delta_mm: float, velocity: Optional[float] = None):
        """Update position by delta_mm relative to current position. Will clamp final position to be within valid range.
        
        Args:
            delta_mm (float): Delta height in millimeters from current position
            velocity (float): Velocity in millimeters per second [0, 200] mm/s, or None to use motor default
        """
        if not self.healthy_state or self.height is None or self.height[0] is None:
            raise Exception("Error: System must be recalibrated.")

        if velocity is not None:
            if velocity < 0 or velocity > 200:
                raise Exception("Error: Velocity is out of range. Must be between 0 and 200 mm/s.")
            self.target_velocity = velocity / MM_IN_CM * ROTATIONS_PER_CM_TRAVELED * (-1 if delta_mm < 0 else 1)
        
        height = max(self.bottom_height_mm, min(self.height[0] + delta_mm, self.top_height_mm))
        #info(f"Updating relative position delta_mm: {delta_mm}, velocity: {velocity}, current height: {self.height[0]}, target height: {height}, target velocity: {self.target_velocity}")
        self.go_to(height)

    @dont_allow_on_fork
    def go_to(self, height_from_bottom: float):
        """Move to target height from bottom.
        
        Args:
            height_from_bottom (float): Target height in millimeters from bottom
        
        Raises:
            Exception: If system not healthy or height is outside valid range
        """
        if not self.healthy_state:
            raise Exception("Error: System must be recalibrated.")
        if height_from_bottom > self.top_height_mm:
            raise Exception(f"Error: Height from bottom {height_from_bottom}mm is greater than maximum supported height {self.top_height_mm}mm.")
        if height_from_bottom < self.bottom_height_mm:
            raise Exception(f"Error: Height from bottom {height_from_bottom}mm is less than minimum supported height {self.bottom_height_mm}mm.")

        height_offset = height_from_bottom - self.reference_position["associated_height"]
        change_in_position = height_offset / MM_IN_CM * ROTATIONS_PER_CM_TRAVELED * STEPS_PER_ROTATION
        target_position = self.reference_position["associated_position"] - change_in_position
        if self.last_command is None or self.last_command[0] != "go_to" or self.last_command[1] != target_position:
            self.last_command = ["go_to", target_position]
            overwrite_file(self.calibration_file_path, ["go_to", str(target_position), str(time.time())])
            self._go_to_raw_position(target_position)

    @dont_allow_on_fork
    def calc_max_height_target_position(self, associated_height: float, associated_position: int):
        """Calculate maximum height target position.
        
        Args:
            associated_height (float): Reference height in millimeters
            associated_position (int): Reference position in motor steps
        
        Returns:
            int: Maximum height target position in motor steps
        """
        height_offset = self.top_height_mm - associated_height
        change_in_position = height_offset / MM_IN_CM * ROTATIONS_PER_CM_TRAVELED * STEPS_PER_ROTATION
        return associated_position - change_in_position

    @dont_allow_on_fork
    def calc_min_height_target_position(self, associated_height: float, associated_position: int):
        """Calculate minimum height target position.
        
        Args:
            associated_height (float): Reference height in millimeters
            associated_position (int): Reference position in motor steps
        
        Returns:
            int: Minimum height target position in motor steps
        """
        height_offset = self.bottom_height_mm - associated_height
        change_in_position = height_offset / MM_IN_CM * ROTATIONS_PER_CM_TRAVELED * STEPS_PER_ROTATION
        return associated_position - change_in_position

    @dont_allow_on_fork
    def shutdown(self):
        """Shutdown motor and stop movement."""
        self.moving = False
        if self._is_enabled():
            if self._ensure_command(
                EZMotionCommands.SET_OPERATION_STATE.value, EZMotionOperationState.SHUTDOWN.value
            ):
                self._record_operation_state(EZMotionOperationState.SHUTDOWN)

    def enable(self):
        """Enable motor operation."""
        if not self._is_enabled():
            if self._ensure_command(
                EZMotionCommands.SET_OPERATION_STATE.value, EZMotionOperationState.ENABLED.value
            ):
                self._record_operation_state(EZMotionOperationState.ENABLED)

    @dont_allow_on_fork
    def transition_to_new_position(self):
        """Transition motor to new position state."""
        set_operation_state = self._ensure_command(EZMotionCommands.SET_OPERATION_STATE.value, EZMotionOperationState.ENABLED.value)
        if not set_operation_state:
            return False
        self.moving = self._ensure_command(
            EZMotionCommands.SET_OPERATION_STATE.value, EZMotionOperationState.TRANSITION_TO_NEW_POSITION.value
        )
        return self.moving

    @dont_allow_on_fork
    def set_control_mode(self, control_mode: EZMotionControlMode):
        """Set motor control mode.
        
        Args:
            control_mode (EZMotionControlMode): Control mode to set
        """
        if self._ensure_command(EZMotionCommands.SET_CONTROL_MODE.value, control_mode.value):
            self.control_mode = control_mode
            return True
        return False

    @dont_allow_on_fork
    def get_status_word(self):
        """Get motor status word.
        
        Returns:
            int | bool: Status word value, or False if read failed
        """
        return self._safe_read(EZMotionCommands.GET_STATUS_WORD.value)

    @dont_allow_on_fork
    def get_current_position(self):
        """Get current motor position.
        
        Returns:
            int | bool: Current position in motor steps, or False if read failed
        """
        return self._safe_read(EZMotionCommands.GET_CURRENT_POSITION.value)

    @dont_allow_on_fork
    def get_current_velocity(self):
        """Get current motor velocity.
        
        Returns:
            int | bool: Current velocity in steps per second, or False if read failed
        """
        return self._safe_read(EZMotionCommands.GET_CURRENT_VELOCITY.value)

    @dont_allow_on_fork
    def health_check(self):
        """Perform health check on motor.
        
        Returns:
            dict: Dictionary with keys: temperature, current, position
        """
        return {
            "temperature": self._safe_read(EZMotionCommands.GET_TEMPERATURE.value),
            "current": self._safe_read(EZMotionCommands.GET_CURRENT_CURRENT.value),
            "position": self.get_current_position(),
        }

    @dont_allow_on_fork
    def get_state(self):
        """Get current motor state.
        
        Returns:
            dict: Dictionary with keys: temperature, velocity, torque, position, height
        """
        return {
            "temperature": self.temperature[0],
            "velocity": self.velocity[0],
            "torque": self.torque[0],
            "position": self.position[0],
            "height": self.height[0]
        }

    @dont_allow_on_fork
    def get_current_position_with_retry(self):
        """Get current position with retry logic.
        
        Returns:
            int: Current position in motor steps
        
        Raises:
            Exception: If position cannot be read after retries
        """
        for i in range(25):
            current_position = self.get_current_position()  # Ensure the position is read
            if current_position is not False:
                return current_position
            time.sleep(0.01)
        raise Exception("Error: Failed to get essential current position.")

    @dont_allow_on_fork
    def get_current_torque(self):
        """Get current motor torque.
        
        Returns:
            int | bool: Current torque value, or False if read failed
        """
        return self._safe_read(EZMotionCommands.GET_CURRENT_TORQUE.value)

    @dont_allow_on_fork
    def on_abort(self):
        warning("TorsoPlatform on_abort")
        """Handle system abort event."""
        if self.aborted:
            return
        self.aborted = True
        if not hasattr(self, "current_position"):
            return
        try:
            self.update_velocity(0, skip_overwrite=True)
        except Exception as e:
            warning(f"Skipping update_velocity: {e}")
        self.shutdown()
        self.network.disconnect()
        self._canopen_stop = True
        if self.canopen_thread is not None:
            self.canopen_thread.join(timeout=1.0)

    @dont_allow_on_fork
    def on_power_event(self, event: PowerEvent):
        """Handle power system events.
        
        Args:
            event (PowerEvent): The power event type (POWER_OFF or POWER_RESTORED)
        """
        if event == PowerEvent.POWER_OFF:
            self.healthy_state = False
        elif event == PowerEvent.POWER_RESTORED:
            self.healthy_state = False
    
    def on_fork(self):
        self.is_forked = True
        self.aborted = True
        self.cfg = {}
        self.motors_map = {}
        self.power = None
        self.print_packet_loss = False
        self.last_sent_velocity = 0
        self.last_sent_acceleration = 0
        self.last_sent_deceleration = 0
        self.last_sent_position = 0
        self.main_motor_id = None
        self.main_motor_model = None
        self.motors_map = {}
        self.calibration_file_path = None
        self.motors_manager = None
        self.health_check_future = None
        self.zero_velocity_count = 0
        self.reference_position = None
        self.network = None
        self.node = None
        self.current_position = None
        self.healthy_state = False
        self.assigned_velocity = 0
        self.last_registered_velocity = (0, -1)
        self.last_step_time = currtime()
        self.operation_state = EZMotionOperationState.SHUTDOWN
        self.last_command = None
        self.moving = False
        self.control_mode = EZMotionControlMode.POSITION
        self.recalibrating = False
        self.height = (None, 0, 0, 0) # (height, timestamp, num_updates)
        self.velocity = (None, 0, 0, 0)
        self.torque = (None, 0, 0, 0)
        self.temperature = (None, 0, 0, 0)
        self.position = (None, 0, 0, 0)



"""Teleop manager: executes high-rate neck (and future teleop) actions from external teleop software."""

from __future__ import annotations

import asyncio
import math
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from .motors.robstride import RobstrideOperationCommand
from .motors.motors_manager import MotorsManager
from feathersdk.utils.logger import info, warning, error

if TYPE_CHECKING:
    from .neck_platform import NeckPlatform
    from .torso_platform import TorsoPlatform
    from .navigation_platform import VelocityNavigationPlanner
    from .manipulation_platform import ManipulationPlatform

# Default (kp, kd) when a motor is not in ARM_OPERATION_GAINS
DEFAULT_ARM_OPERATION_KP = 30.0
DEFAULT_ARM_OPERATION_KD = 5.0

# motor_name -> (kp, kd) for arm operation commands. Edit here if config loading is not used.
ARM_OPERATION_GAINS: Dict[str, Tuple[float, float]] = {
    "SpR": (500.0, 5.0),
    "SrR": (500.0, 5.0),
    "SwR": (30.0, 5.0),
    "EpR": (300.0, 5.0),
    "WwR": (30.0, 5.0),
    "WpR": (30.0, 5.0),
    "WrR": (30.0, 5.0),
    "SpL": (500.0, 5.0),
    "SrL": (500.0, 5.0),
    "SwL": (30.0, 5.0),
    "EpL": (300.0, 5.0),
    "WwL": (30.0, 5.0),
    "WpL": (30.0, 5.0),
    "WrL": (30.0, 5.0),
}

# If True, on_arm_action uses execute_arm_go_to (trajectory); if False, uses execute_operation_commands.
USE_GO_TO = False


@dataclass
class NeckAction:
    """Neck command - arrives at ~120Hz (future feature).

    Teleop software sends this; the teleop manager forwards it to the neck platform.
    """

    joint_names: List[str]   # e.g. ["neck_pan", "neck_tilt"]
    positions: List[float]   # radians
    velocities: List[float]  # rad/s
    efforts: List[float]     # Target efforts in Nm (will be zeros)
    timestamp: float


@dataclass
class ElevatorAction:
    """Elevator command - arrives at ~30Hz continuously.

    Teleop software sends this; the teleop manager forwards it to the torso platform.
    """

    velocity: float   # Normalized z velocity -1.0 to 1.0 (0=stop, positive=up, negative=down; can be binary)
    timestamp: float


@dataclass
class LocomotionAction:
    """Locomotion command - arrives at ~30Hz continuously (even when zero).

    Teleop software sends this; the teleop manager forwards it to the nav planner.
    """

    linear_x: float   # Forward/backward, normalized -1.0 to 1.0
    linear_y: float  # Left/right, normalized -1.0 to 1.0
    angular_z: float  # Rotation, normalized -1.0 to 1.0
    timestamp: float


@dataclass
class ArmAction:
    """Arm command - arrives at ~120Hz continuously.

    Teleop software sends this; the teleop manager forwards it to the arm motors
    via operation commands (target angle, velocity, effort, kp/kd).
    """

    joint_names: List[str]   # e.g. ["SpL", "SrL", "SwL", ...]
    positions: List[float]   # Target positions in radians
    velocities: List[float]  # Target velocities in rad/s (often zeros)
    efforts: List[float]    # Target efforts in Nm (often zeros)
    timestamp: float        # When command was received (seconds since epoch)


class TeleopManager:
    """
    All execute_* methods are non-blocking: they set targets or queue commands and return
    immediately. Actual motion is performed by each platform's step loop or background thread.
    """

    def __init__(
        self,
        neck: Optional["NeckPlatform"] = None,
        torso: Optional["TorsoPlatform"] = None,
        nav_planner: Optional["VelocityNavigationPlanner"] = None,
        manipulation_platform: Optional["ManipulationPlatform"] = None,
        motors_manager: Optional[MotorsManager] = None,
        config: Optional[dict] = None,
    ):
        """Initialize with the neck, torso, nav planner, and/or motors manager to control.
        config: optional dict for torso_max_velocity_mm_s, elevator_deadzone, locomotion_deadzone,
            locomotion_max_linear_m_s, locomotion_max_rotation_deg_s. Arm gains use module-level
            ARM_OPERATION_GAINS and DEFAULT_ARM_OPERATION_KP/KD.
        """
        cfg = config if isinstance(config, dict) else {}
        self._neck = neck
        self._torso = torso
        self._nav_planner = nav_planner
        self._manipulation_platform = manipulation_platform
        self._motors_manager = motors_manager
        self._default_arm_operation_kp = DEFAULT_ARM_OPERATION_KP
        self._default_arm_operation_kd = DEFAULT_ARM_OPERATION_KD
        self.arm_operation_gains = dict(ARM_OPERATION_GAINS)
        self._torso_max_velocity_mm_s = float(cfg.get("torso_max_velocity_mm_s", 50.0))
        self._elevator_deadzone = float(cfg.get("elevator_deadzone", 0.2))
        self._locomotion_deadzone = float(cfg.get("locomotion_deadzone", 0.2))
        self._locomotion_max_linear_m_s = float(cfg.get("locomotion_max_linear_m_s", 0.3))
        self._locomotion_max_rotation_deg_s = float(cfg.get("locomotion_max_rotation_deg_s", 100.0))
        # Cache last-sent values to avoid redundant downstream calls.
        self._last_torso_velocity: Optional[float] = None
        self._last_neck: Optional[Tuple[Optional[float], Optional[float]]] = None
        self._last_locomotion: Optional[Tuple[str, float]] = None  # ('stop'|'linear_x'|'linear_y'|'angular_z', value)
        self._last_arm: Optional[Tuple[Tuple[str, ...], Tuple[float, ...], Tuple[float, ...], Tuple[float, ...]]] = None
        self._last_open_left: Optional[float] = None
        self._last_open_right: Optional[float] = None
        self._last_grip_left: Optional[float] = None
        self._last_grip_right: Optional[float] = None

        # if self._manipulation_platform is not None:
        #     warning(f"Enabling arms")
        #     self._manipulation_platform.right_arm.enable()
        #     self._manipulation_platform.left_arm.enable()
        #     kp_by_motor = {m: ARM_OPERATION_GAINS[m][0] for m in ARM_OPERATION_GAINS}
        #     kd_by_motor = {m: ARM_OPERATION_GAINS[m][1] for m in ARM_OPERATION_GAINS}
        #     self._manipulation_platform.right_arm.freeze_current_arm_state(kp_by_motor, kd_by_motor)
        #     self._manipulation_platform.left_arm.freeze_current_arm_state(kp_by_motor, kd_by_motor)
        #     time.sleep(1)
        #     self._manipulation_platform.right_arm.home()
        #     self._manipulation_platform.left_arm.home()
       

    def on_open_trigger_change(self, is_right: bool = True, percentage: float = 0.0) -> None:
        """Handle open-trigger change from teleop. Prints only when value changed."""
        pct = percentage
        if is_right:
            if self._last_open_right == pct:
                return
            self._last_open_right = pct
        else:
            if self._last_open_left == pct:
                return
            self._last_open_left = pct
        side = "right" if is_right else "left"
        print(f"open_trigger {side}: {pct}")
        warning(f"open_trigger {side}: {pct}")

    def on_grip_trigger_change(self, is_right: bool = True, percentage: float = 0.0) -> None:
        """Handle grip-trigger change from teleop. Prints only when value changed."""
        pct = percentage
        if is_right:
            if self._last_grip_right == pct:
                return
            self._last_grip_right = pct
        else:
            if self._last_grip_left == pct:
                return
            self._last_grip_left = pct
        side = "right" if is_right else "left"
        print(f"grip_trigger {side}: {pct}")

    def on_y_press(self) -> None:
        """Handle Y button press from teleop. Override or hook as needed."""
        if self._manipulation_platform is not None:
            self._motors_manager.trigger_compliance_mode("right_arm", "SpR")
            self._motors_manager.trigger_compliance_mode("left_arm", "SpL")
        if self._torso is not None:
            self._torso.update_velocity(0)
        if self._nav_planner is not None:
            self._nav_planner.set_speed(0)
        print("Y pressed")

    def start_dataset(self) -> None:
        """Start dataset recording. Override or hook as needed."""
        print("start_dataset")

    def stop_dataset(self) -> None:
        """Stop dataset recording. Override or hook as needed."""
        print("stop_dataset")

    def execute_neck_recommended_action(self, action: NeckAction) -> None:
        """Apply a single NeckAction to the neck platform. Safe to call at ~120Hz.
        """
        warning(f"{action.timestamp} execute_neck_recommended_action: {action}")
        neck = self._neck
        if neck is None:
            return

        pitch_rad: Optional[float] = None
        yaw_rad: Optional[float] = None
        names = action.joint_names
        positions = action.positions
        n = len(names)
        if n != len(positions):
            return

        for i in range(n):
            name = names[i]
            pos = positions[i]
            if name == neck.PITCH_MOTOR_NAME:
                pitch_rad = pos
            elif name == neck.YAW_MOTOR_NAME:
                yaw_rad = pos

        if pitch_rad is None and yaw_rad is None:
            return

        pitch_deg = math.degrees(pitch_rad) if pitch_rad is not None else None
        yaw_deg = math.degrees(yaw_rad) if yaw_rad is not None else None
        key = (pitch_deg, yaw_deg)
        if self._last_neck == key:
            return
        self._last_neck = key
        neck.go_to(pitch_in_degrees=pitch_deg, yaw_in_degrees=yaw_deg)

    def execute_torso_recommended_action(self, action: ElevatorAction) -> None:
        """Apply a single ElevatorAction to the torso platform. Safe to call at ~30Hz.
        """
        warning(f"{action.timestamp} execute_torso_recommended_action: {action}")
        torso = self._torso
        if torso is None:
            return
        v = max(-1.0, min(1.0, float(action.velocity)))
        # Deadzone: treat stick drift as stop so torso does not creep
        if -self._elevator_deadzone <= v <= self._elevator_deadzone:
            v = 0.0
        mm_per_second = v * self._torso_max_velocity_mm_s
        if self._last_torso_velocity == mm_per_second:
            return
        self._last_torso_velocity = mm_per_second
        try:
            torso.update_velocity(mm_per_second)
        except Exception:
            pass  # e.g. not healthy, or at limits; torso already no-ops at limits
        

    def execute_arm_recommended_action(self, action: ArmAction) -> None:
        """Apply a single ArmAction to arm motors via operation commands. Safe to call at ~120Hz.
        """
        if USE_GO_TO:
            self.execute_arm_go_to(action) # 120Hz
        else:
            self.execute_operation_commands(action)

    def execute_operation_commands(self, action: ArmAction) -> None:
        mm = self._motors_manager
        if mm is None:
            return
        names = action.joint_names
        positions = action.positions
        velocities = action.velocities
        efforts = action.efforts

        n = len(names)
        if n != len(positions):
            error(f"{action.timestamp} execute_arm_recommended_action: len(positions) != len(names)")
            return
        key = (tuple(names), tuple(positions), tuple(velocities), tuple(efforts))
        if self._last_arm == key:
            return
        self._last_arm = key
        default_gains = (self._default_arm_operation_kp, self._default_arm_operation_kd)
        gains = self.arm_operation_gains
        commands: List[RobstrideOperationCommand] = []
        for i in range(n):
            motor_name = names[i]
            try:
                motor = mm.get_motor(motor_name)
            except Exception as e:
                error(f"error getting motor {motor_name}: {e}")
            kp, kd = gains.get(motor_name, default_gains)

            warning(f"SENDING operation command to {motor_name}: {motor.convert_calibrated_angle_to_angle(positions[i])} kp: {kp} kd: {kd}")
            cmd = RobstrideOperationCommand(
                motor,
                0,
                motor.convert_calibrated_angle_to_angle(positions[i]),
                velocities[i],
                kp,
                kd,
            )
            commands.append(cmd)
        if commands:
            try:
                mm.send_operation_commands(commands)
            except Exception as e:
                error(f"{action.timestamp} execute_arm_recommended_action: {e}")
                pass

    def execute_arm_go_to(self, action: ArmAction, total_time: float = 1.0) -> None:
        """Apply ArmAction to arms using arm.go_to (trajectory). Non-blocking.

        Builds motor name -> target position from action.joint_names and action.positions
        (radians / calibrated angle), then calls left_arm.go_to and right_arm.go_to.
        """
        warning(f"execute_arm_go_to: {action}")
        platform = self._manipulation_platform
        if platform is None:
            return
        names = action.joint_names
        positions = action.positions
        if len(names) != len(positions):
            error(f"execute_arm_go_to: len(joint_names) != len(positions)")
            return
        motors_targets = dict(zip(names, positions))
        # Use each arm's motor names as source of truth (arm_system.motors is Dict[str, Motor]).
        right_targets = {m: motors_targets[m] for m in platform.right_arm.motors if m in motors_targets}
        left_targets = {m: motors_targets[m] for m in platform.left_arm.motors if m in motors_targets}
        if not right_targets and not left_targets:
            warning(
                f"execute_arm_go_to: no targets for any arm. "
                f"right_arm.motors keys={list(platform.right_arm.motors.keys())!r}, "
                f"left_arm.motors keys={list(platform.left_arm.motors.keys())!r}, "
                f"action joint_names={names!r}"
            )
        if right_targets:
            try:
                warning(f"execute_arm_go_to right_arm: {right_targets} {total_time}")
                platform.right_arm.go_to(right_targets, total_time)
            except Exception as e:
                error(f"execute_arm_go_to right_arm: {e}")
        if left_targets:
            try:
                warning(f"execute_arm_go_to left_arm: {left_targets} {total_time}")
                platform.left_arm.go_to(left_targets, total_time)
            except Exception as e:
                error(f"execute_arm_go_to left_arm: {e}")

    def execute_base_recommended_action(self, action: LocomotionAction) -> None:
        """Apply a single LocomotionAction to the nav planner. Safe to call at ~30Hz.
        """
        warning(f"{action.timestamp} execute_base_recommended_action: {action}")
        nav = self._nav_planner
        if nav is None:
            return
        dx = max(-1.0, min(1.0, float(action.linear_x)))
        dy = max(-1.0, min(1.0, float(action.linear_y)))
        dz = max(-1.0, min(1.0, float(action.angular_z)))
        # Apply deadzone: treat as zero inside [-locomotion_deadzone, locomotion_deadzone]
        def above_deadzone(v: float) -> float:
            if -self._locomotion_deadzone <= v <= self._locomotion_deadzone:
                return 0.0
            return v
        ax = above_deadzone(dx)
        ay = above_deadzone(dy)
        az = above_deadzone(dz)
        mag_x = abs(ax)
        mag_y = abs(ay)
        mag_z = abs(az)
        if mag_x == 0 and mag_y == 0 and mag_z == 0:
            cmd = ("stop", 0.0)
        elif mag_x >= mag_y and mag_x >= mag_z:
            cmd = ("linear_x", ax * self._locomotion_max_linear_m_s)
        elif mag_y >= mag_z:
            cmd = ("linear_y", ay * self._locomotion_max_linear_m_s)
        else:
            cmd = ("angular_z", -az * self._locomotion_max_rotation_deg_s)
        if self._last_locomotion == cmd:
            return
        self._last_locomotion = cmd
        try:
            if cmd[0] == "stop":
                nav.set_speed(0)
            elif cmd[0] == "linear_x":
                nav.go_to(0, cmd[1])
            elif cmd[0] == "linear_y":
                nav.go_to(90, cmd[1])
            else:
                nav.rotate_in_place(cmd[1])
        except Exception:
            pass
    
    
    def arm_observation(self) -> Dict[str, dict]:
        """Read current state of arms. Returns motor_name -> {temp, torque, angle, velocity}."""
        platform = self._manipulation_platform
        if platform is None:
            return {}
        out: Dict[str, dict] = {}
        out.update(platform.right_arm.get_state())
        out.update(platform.left_arm.get_state())
        return out

    def handle_dataset_start(self, dataset_id: str) -> None:
        print(f"Dataset {dataset_id} started")
        warning(f"Dataset {dataset_id} started")

    def handle_dataset_stop(self, dataset_id: str) -> None:
        print(f"Dataset {dataset_id} stopped")
        warning(f"Dataset {dataset_id} stopped")

    def on_y_pressed(self) -> None:
        print("Y pressed, emergency stop")
        warning("Y pressed, emergency stop")